"""Covariance matrix."""
import numpy

from .variance import Var
from .. import distributions, poly as polynomials


def Cov(poly, dist=None, **kws):
    """
    Covariance matrix, or 2rd order statistics.

    Args:
        poly (Poly, Dist) :
            Input to take covariance on. Must have `len(poly)>=2`.
        dist (Dist) :
            Defines the space the covariance is taken on.  It is ignored if
            `poly` is a distribution.

    Returns:
        (numpy.ndarray):
            Covariance matrix with shape ``poly.shape+poly.shape``.

    Examples:
        >>> dist = chaospy.MvNormal([0, 0], [[2, .5], [.5, 1]])
        >>> print(chaospy.Cov(dist))
        [[2.  0.5]
         [0.5 1. ]]
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> print(chaospy.Cov(poly, dist))
        [[  0.    0.    0.    0. ]
         [  0.    2.    0.5   0. ]
         [  0.    0.5   1.    0. ]
         [  0.    0.    0.  225. ]]
    """
    if not isinstance(poly, (distributions.Dist, polynomials.Poly)):
        poly = polynomials.Poly(poly)

    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    if not poly.shape:
        return Var(poly, dist).reshape(1, 1)

    dim = len(dist)
    shape = poly.shape
    poly = polynomials.flatten(poly)
    keys = poly.keys
    N = len(keys)
    A = poly.A
    keys1 = numpy.array(keys).T
    if dim==1:
        keys1 = keys1[0]
        keys2 = sum(numpy.meshgrid(keys, keys))
    else:
        keys2 = numpy.empty((dim, N, N))
        for i in range(N):
            for j in range(N):
                keys2[:, i, j] = keys1[:, i]+keys1[:, j]

    m1 = dist.mom(keys1, **kws)
    m2 = dist.mom(keys2, **kws)
    mom = m2-numpy.outer(m1, m1)

    out = numpy.zeros((len(poly), len(poly)))
    for i in range(len(keys)):
        a = A[keys[i]]
        out += numpy.outer(a, a)*mom[i, i]
        for j in range(i+1, len(keys)):
            b = A[keys[j]]
            ab = numpy.outer(a, b)
            out += (ab+ab.T)*mom[i, j]

    out = numpy.reshape(out, shape+shape)
    return out



"""Function for handling stochastic dependencies."""
from .common import DependencyError


def sorted_dependencies(dist, reverse=False):
    """
    Extract all underlying dependencies from a distribution sorted
    topologically.

    Uses depth-first algorithm. See more here:

    Args:
        dist (Dist):
            Distribution to extract dependencies from.
        reverse (bool):
            If True, place dependencies in reverse order.

    Returns:
        dependencies (List[Dist]):
            All distribution that ``dist`` is dependent on, sorted
            topologically, including itself.

    Examples:
        >>> dist1 = chaospy.Uniform()
        >>> dist2 = chaospy.Normal(dist1)
        >>> print(sorted_dependencies(dist1))
        [Uniform(lower=0, upper=1), Mul(uniform(), 0.5), uniform()]
        >>> print(sorted_dependencies(dist2)) # doctest: +NORMALIZE_WHITESPACE
        [Normal(mu=Uniform(lower=0, upper=1), sigma=1),
         Uniform(lower=0, upper=1),
         Mul(uniform(), 0.5),
         uniform(),
         Mul(normal(), 1),
         normal()]
        >>> dist1 in sorted_dependencies(dist2)
        True
        >>> dist2 in sorted_dependencies(dist1)
        False

    Raises:
        DependencyError:
            If the dependency DAG is cyclic, dependency resolution is not
            possible.

    See also:
        Depth-first algorithm section:
        https://en.wikipedia.org/wiki/Topological_sorting
    """
    from .. import baseclass

    collection = [dist]

    # create DAG as list of nodes and edges:
    nodes = [dist]
    edges = []
    pool = [dist]
    while pool:
        dist = pool.pop()
        for key in sorted(dist.prm):
            value = dist.prm[key]
            if not isinstance(value, baseclass.Dist):
                continue
            if (dist, value) not in edges:
                edges.append((dist, value))
            if value not in nodes:
                nodes.append(value)
                pool.append(value)

    # temporary stores used by depth first algorith.
    permanent_marks = set()
    temporary_marks = set()

    def visit(node):
        """Depth-first topological sort algorithm."""
        if node in permanent_marks:
            return
        if node in temporary_marks:
            raise DependencyError("cycles in dependency structure.")

        nodes.remove(node)
        temporary_marks.add(node)

        for node1, node2 in edges:
            if node1 is node:
                visit(node2)

        temporary_marks.remove(node)
        permanent_marks.add(node)
        pool.append(node)

    # kickstart algorithm.
    while nodes:
        node = nodes[0]
        visit(node)

    if not reverse:
        pool = list(reversed(pool))

    return pool


def get_dependencies(*distributions):
    """
    Get underlying dependencies that are shared between distributions.

    If more than two distributions are provided, any pair-wise dependency
    between any two distributions are included, implying that an empty set is
    returned if and only if the distributions are i.i.d.

    Args:
        distributions:
            Distributions to check for dependencies.

    Returns:
        dependencies (List[Dist]):
            Distributions dependency shared at least between at least one pair
            from ``distributions``.

    Examples:
        >>> dist1 = chaospy.Uniform(1, 2)
        >>> dist2 = chaospy.Uniform(1, 2) * dist1
        >>> dist3 = chaospy.Uniform(3, 5)
        >>> print(chaospy.get_dependencies(dist1, dist2))
        [uniform(), Mul(uniform(), 0.5), Uniform(lower=1, upper=2)]
        >>> print(chaospy.get_dependencies(dist1, dist3))
        []
        >>> print(chaospy.get_dependencies(dist2, dist3))
        []
        >>> print(chaospy.get_dependencies(dist1, dist2, dist3))
        [uniform(), Mul(uniform(), 0.5), Uniform(lower=1, upper=2)]
    """
    from .. import baseclass
    distributions = [
        sorted_dependencies(dist) for dist in distributions
        if isinstance(dist, baseclass.Dist)
    ]

    dependencies = list()
    for idx, dist1 in enumerate(distributions):
        for dist2 in distributions[idx+1:]:
            dependencies.extend([dist for dist in dist1 if dist in dist2])

    return sorted(dependencies)



"""Nakagami-m distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class nakagami(Dist):
    """Nakagami-m distribution."""

    def __init__(self, nu):
        Dist.__init__(self, nu=nu)

    def _pdf(self, x, nu):
        return 2*nu**nu/special.gamma(nu)*(x**(2*nu-1.0))*numpy.exp(-nu*x*x)

    def _cdf(self, x, nu):
        return special.gammainc(nu,nu*x*x)

    def _ppf(self, q, nu):
        return numpy.sqrt(1.0/nu*special.gammaincinv(nu, q))

    def _bnd(self, x, nu):
        return 0.0, self._ppf(1-1e-10, nu)


class Nakagami(Add):
    """
    Nakagami-m distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Nakagami(2, 2, 2)
        >>> print(distribution)
        Nakagami(scale=2, shape=2, shift=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [3.284  3.6592 4.0111 4.4472]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.4642 0.5766 0.5383 0.3669]
        >>> print(numpy.around(distribution.sample(4), 4))
        [4.1137 3.076  5.0824 3.8012]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.88
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[4.1568 4.4181 4.6622]
         [0.4657 0.8824 1.2706]]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, nakagami(shape)*scale, shift)



"""Frank copula."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist

class frank(Archimedean):
    """Frank copula."""

    def __init__(self, length, theta, eps=1e-6):
        assert theta != 0
        self.length = length
        Dist.__init__(self, th=float(theta), eps=eps)

    def __len__(self):
        return self.length

    def gen(self, x, th):
        return -numpy.log((numpy.e**(-th*x)-1)/(numpy.e**-th-1))

    def igen(self, q, th):
        return -numpy.log(1+numpy.e**-q*(numpy.e**-th-1))/th

class Frank(Copula):
    """
    Frank copula.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Uniform(), 2)
        >>> distribution = chaospy.Frank(distribution)
        >>> print(distribution)
        Frank(Iid(Uniform(lower=0, upper=1), 2), theta=1.0)
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.2101 0.2539 0.3032]
          [0.4391 0.5    0.5609]
          [0.6968 0.7462 0.7899]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[1.1454 1.0054 0.9031]
         [1.0358 1.0207 1.0358]
         [0.9031 1.0055 1.1454]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.8854 0.1587 0.0646 0.395 ]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        0.1804
    """

    def __init__(self, dist, theta=1., eps=1e-4):
        """
        Args:
            dist (Dist) : The Distribution to wrap
            theta (float) : Copula parameter
        """
        self._repr = {"theta": theta}
        return Copula.__init__(self, dist=dist, trans=frank(len(dist), theta, eps))



"""Logarithm with base Euler's constant."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Log(Dist):
    """
    Logarithm with base Euler's constant.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Log(chaospy.Uniform(1, 2))
        >>> print(distribution)
        Log(Uniform(lower=1, upper=2))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.1823 0.3365 0.47   0.5878]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.2 1.4 1.6 1.8]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.5029 0.1089 0.668  0.3935]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.3863
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.3863 0.3389 0.3454]
         [1.     0.0391 0.0321]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range() > 0)
        Dist.__init__(self, dist=dist)

    def _pdf(self, xloc, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.e**xloc, cache=cache)*numpy.e**xloc

    def _cdf(self, xloc, dist, cache):
        """Cumulative distribution function."""
        return evaluation.evaluate_forward(dist, numpy.e**xloc, cache=cache)

    def _ppf(self, q, dist, cache):
        """Point percentile function."""
        return numpy.log(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, xloc, dist, cache):
        """Distribution bounds."""
        return numpy.log(evaluation.evaluate_bound(
            dist, numpy.e**xloc, cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.log(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.e**dist
        return self



"""Sinus Hyperbolic."""
import numpy
from ..baseclass import Dist
from .. import evaluation, approximation


class Sinh(Dist):
    """
    Sinus Hyperbolic.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Sinh(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Sinh(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.2013 0.4108 0.6367 0.8881]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.9803 0.925  0.8436 0.7477]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.7011 0.1153 1.0999 0.5011]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.5431
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.5431 0.5971 0.5885]
         [1.     0.1118 0.0927]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.arcsinh(x), cache=cache)/numpy.sqrt(1+x*x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.arcsinh(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.sinh(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.sinh(evaluation.evaluate_bound(
            dist, numpy.arcsinh(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"



"""Tukey-lambda distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class tukey_lambda(Dist):
    """Tukey-lambda distribution."""

    def __init__(self, lam):
        Dist.__init__(self, lam=lam)

    def _pdf(self, x, lam):
        lam = numpy.zeros(x.shape) + lam
        output = numpy.zeros(x.shape)
        indices = (lam <= 0) | (numpy.abs(x)*lam < 1)
        lam = lam[indices]
        Fx = special.tklmbda(x[indices], lam)
        Px = 1/(Fx**(lam-1.0) + ((1-Fx))**(lam-1.0))
        output[indices] = Px
        return output

    def _cdf(self, x, lam):
        return special.tklmbda(x, lam)

    def _ppf(self, q, lam):
        output = numpy.zeros(q.shape)
        lam = numpy.zeros(q.shape) + lam
        indices = lam != 0
        q_ = q[indices]
        lam_ = lam[indices]
        output[indices] = (q_**lam_ - (1-q_)**lam_)/lam_
        q_ = q[~indices]
        output[~indices] = numpy.log(q_/(1-q_))
        return output

    def _bnd(self, x, lam):
        return (
            self._ppf(numpy.array(1e-10), lam),
            self._ppf(numpy.array(1-1e-10), lam),
        )


class TukeyLambda(Add):
    """
    Tukey-lambda distribution.

    Args:
        lam (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.TukeyLambda(0, 2, 2)
        >>> print(distribution)
        TukeyLambda(scale=2, shape=0, shift=2)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [-1.2189  0.6137  2.      3.3863  5.2189]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0694 0.1111 0.125  0.1111 0.0694]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 3.2697 -2.0812  7.9008  1.8575]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 2.      2.      2.    ]
         [13.1595 42.1102 91.3601]]
    """

    def __init__(self, shape=0, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=tukey_lambda(shape)*scale, right=shift)


Tukeylambda = deprecation_warning(TukeyLambda, "Tukeylambda")



r"""
A cumulative distribution function of an independent multivariate random
variable can be made dependent through a copula as follows:

.. math::
    F_{Q_0,\dots,Q_{D-1}} (q_0,\dots,q_{D-1}) =
    C(F_{Q_0}(q_0), \dots, F_{Q_{D-1}}(q_{D-1}))

where :math:`C` is the copula function, and :math:`F_{Q_i}` are marginal
distribution functions.  One of the more popular classes of copulas is the
Archimedean copulas.
.. \cite{sklar_random_1996}.
They are defined as follows:

.. math::
    C(u_1,\dots,u_n) =
    \phi^{[-1]} (\phi(u_1)+\dots+\phi(u_n)),

where :math:`\phi` is a generator and :math:`\phi^{[-1]}` is its
pseudo-inverse. Support for Archimedean copulas in `chaospy` is possible
through reformulation of the Rosenblatt transformation.  In two dimension, this
reformulation is as follows:

.. math::

    F_{U_0}(u_0) = \frac{C(u_0,1)}{C(1,1)}

    F_{U_1\mid U_0}(u_1\mid u_0) =
    \frac{\tfrac{\partial}{\partial u_0}
    C(u_0,u_1)}{\tfrac{\partial}{\partial u_0} C(u_0,1)}

This definition can also be generalized in to multiple variables using the
formula provided by Nelsen 1999.
.. cite:: nelsen_introduction_1999

The definition of the Rosenblatt transform can require multiple
differentiations.  An analytical formulation is usually not feasible, so the
expressions are estimated using difference scheme similar to the one outlined
for probability density function defined in :ref:`distributions`. The accurate
might therefore be affected.

Since copulas are meant as a replacement for Rosenblatt
transformation, it is usually assumed that the distribution it is
used on is stochastically independent.
However in the definition of a copula does not actually require it, and sine
the Rosenblatt transformation allows for it, multiple copulas can be stacked
together in `chaospy`.
"""
import numpy

from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation


class Copula(Dist):

    def __init__(self, dist, trans):
        r"""
        Args:
            dist (Dist) : Distribution to wrap the copula around.
            trans (Dist) : The copula wrapper `[0,1]^D \into [0,1]^D`.
        """
        Dist.__init__(self, dist=dist, trans=trans)

    def _cdf(self, x, dist, trans, cache):
        output = evaluation.evaluate_forward(dist, x, cache=cache)
        output = evaluation.evaluate_forward(trans, output, cache=cache)
        return output

    def _bnd(self, x, dist, trans, cache):
        return evaluation.evaluate_bound(dist, x, cache=cache)

    def _ppf(self, qloc, dist, trans, cache):
        qloc = evaluation.evaluate_inverse(trans, qloc, cache=cache)
        xloc = evaluation.evaluate_inverse(dist, qloc, cache=cache)
        return xloc

    def _pdf(self, x, dist, trans, cache):
        density = evaluation.evaluate_density(dist, x, cache=cache.copy())
        return evaluation.evaluate_density(
            trans, evaluation.evaluate_forward(
                dist, x, cache=cache), cache=cache)*density

    def _mom(self, x, dist, trans, cache):
        raise StochasticallyDependentError(
            "Joint distribution with dependencies not supported.")

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        args = [key + "=" + str(self._repr[key]) for key in sorted(self._repr)]
        return (self.__class__.__name__ + "(" + str(self.prm["dist"]) +
                ", " + ", ".join(args) + ")")


class Archimedean(Dist):
    """
    Archimedean copula superclass.

    Subset this to generate an Archimedean copula.
    """

    def _ppf(self, x, th, eps):

        for i in range(1, len(x)):

            q = x[:i+1].copy()
            lo, up = 0, 1
            dq = numpy.zeros(i+1)
            dq[i] = eps
            flo, fup = -q[i],1-q[i]

            for iteration in range(1, 10):

                fq = self._diff(q[:i+1], th, eps)
                dfq = self._diff((q[:i+1].T+dq).T, th, eps)
                dfq = (dfq-fq)/eps
                dfq = numpy.where(dfq==0, numpy.inf, dfq)

                fq = fq-x[i]
                if not numpy.any(numpy.abs(fq)>eps):
                    break

                # reduce boundaries
                flo = numpy.where(fq<=0, fq, flo)
                lo = numpy.where(fq<=0, q[i], lo)

                fup = numpy.where(fq>=0, fq, fup)
                up = numpy.where(fq>=0, q[i], up)

                # Newton increment
                qdq = q[i]-fq/dfq

                # if new val on interior use Newton
                # else binary search
                q[i] = numpy.where((qdq<up)*(qdq>lo),
                        qdq, .5*(up+lo))

            x[i] = q[i]
        return x


    def _cdf(self, x, th, eps):
        out = numpy.zeros(x.shape)
        out[0] = x[0]
        for i in range(1,len(x)):
            out[i][x[i]==1] = 1
            out[i] = self._diff(x[:i+1], th, eps)

        return out

    def _pdf(self, x, th, eps):
        out = numpy.ones(x.shape)
        sign = 1-2*(x>.5)
        for i in range(1, len(x)):
            x[i] += eps*sign[i]
            out[i] = self._diff(x[:i+1], th, eps)
            x[i] -= eps*sign[i]
            out[i] -= self._diff(x[:i+1], th, eps)
            out[i] /= eps

        out = abs(out)
        return out

    def _diff(self, x, th, eps):
        """
        Differentiation function.

        Numerical approximation of a Rosenblatt transformation created from
        copula formulation.
        """
        foo = lambda y: self.igen(numpy.sum(self.gen(y, th), 0), th)

        out1 = out2 = 0.
        sign = 1 - 2*(x > .5).T
        for I in numpy.ndindex(*((2,)*(len(x)-1)+(1,))):

            eps_ = numpy.array(I)*eps
            x_ = (x.T + sign*eps_).T
            out1 += (-1)**sum(I)*foo(x_)

            x_[-1] = 1
            out2 += (-1)**sum(I)*foo(x_)

        out = out1/out2
        return out


    def _bnd(self, **prm):
        return 0, 1



"""
Create `antithetic variates`_ from variables on the unit hyper-cube.

In statistics, the antithetic variates method is a variance reduction technique
used in Monte Carlo methods. Considering that the error reduction in the
simulated signal (using Monte Carlo methods) has a square root convergence,
a very large number of sample paths is required to obtain an accurate result.
The antithetic variates method reduces the variance of the simulation results.

Antithetic variate can be accessed as a flag ``antithetic`` in the method
``Dist.sample`` It can either be set to ``True``, for activation, or as an
array of boolean values, which implies it will be used as the flag ``axes`` in
the examples below.

Example usage
-------------

Creating antithetic variates can be done directly from each distribution by
using the ``antithetic`` flag::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> samples = distribution.sample(6, antithetic=True)

Antithetic variates contains compliment values of itself::

    >>> print(numpy.around(samples, 4))
    [0.6536 0.3464 0.115  0.885  0.9503 0.0497]
    >>> print(numpy.around(1-samples, 4))
    [0.3464 0.6536 0.885  0.115  0.0497 0.9503]

Antithetic variates can also be used in multiple dimensions::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> samples = distribution.sample(6, antithetic=True)
    >>> print(numpy.around(samples, 4))
    [[0.0407 0.9593 0.0407 0.9593 0.3972 0.6028]
     [0.8417 0.8417 0.1583 0.1583 0.2071 0.2071]]
    >>> print(numpy.around(1-samples, 4))
    [[0.9593 0.0407 0.9593 0.0407 0.6028 0.3972]
     [0.1583 0.1583 0.8417 0.8417 0.7929 0.7929]]

Lastly, it is also possible to select which axes should be included when
applying the variate by passing a bool array. For axes that are "false", the
value is frozen in place::

    >>> samples = distribution.sample(6, antithetic=[True, False])
    >>> print(numpy.around(samples, 4))
    [[0.3922 0.6078 0.1823 0.8177 0.7435 0.2565]
     [0.0696 0.0696 0.8853 0.8853 0.9526 0.9526]]
    >>> print(numpy.around(1-samples, 4))
    [[0.6078 0.3922 0.8177 0.1823 0.2565 0.7435]
     [0.9304 0.9304 0.1147 0.1147 0.0474 0.0474]]
    >>> samples = distribution.sample(6, antithetic=[False, True])
    >>> print(numpy.around(samples, 4))
    [[0.9311 0.9311 0.4154 0.4154 0.029  0.029 ]
     [0.982  0.018  0.3396 0.6604 0.7067 0.2933]]
    >>> print(numpy.around(1-samples, 4))
    [[0.0689 0.0689 0.5846 0.5846 0.971  0.971 ]
     [0.018  0.982  0.6604 0.3396 0.2933 0.7067]]

.. _antithetic variates: https://en.wikipedia.org/wiki/Antithetic_variates
"""
import numpy


def create_antithetic_variates(samples, axes=()):
    """
    Generate antithetic variables.

    Args:
        samples (numpy.ndarray):
            The samples, assumed to be on the [0, 1]^D hyper-cube, to be
            reflected.
        axes (tuple):
            Boolean array of which axes to reflect. If This to limit the number
            of points created in higher dimensions by reflecting all axes at
            once.

    Returns (numpy.ndarray):
        Same as ``samples``, but with samples internally reflected. roughly
        equivalent to ``numpy.vstack([samples, 1-samples])`` in one dimensions.
    """
    samples = numpy.asfarray(samples)
    assert numpy.all(samples <= 1) and numpy.all(samples >= 0), (
        "all samples assumed on interval [0, 1].")
    if len(samples.shape) == 1:
        samples = samples.reshape(1, -1)
    inverse_samples = 1-samples
    dims = len(samples)

    if not len(axes):
        axes = (True,)
    axes = numpy.asarray(axes, dtype=bool).flatten()

    indices = {tuple(axes*idx) for idx in numpy.ndindex((2,)*dims)}
    indices = sorted(indices, reverse=True)
    indices = sorted(indices, key=lambda idx: sum(idx))
    out = [numpy.where(idx, inverse_samples.T, samples.T).T for idx in indices]
    out = numpy.dstack(out).reshape(dims, -1)
    return out



# pylint: disable=wildcard-import
"""
One of the backbone of any uncertainty quantification is a collection of
probability distributions, and ``chaospy`` is no exception. For example, to
create a Gaussian random variable::

    >>> distribution = chaospy.Normal(mu=2, sigma=2)

The syntax for using distribution is here very similar to the syntax used in
``scipy.dist``. For example, to create values from the *probability density
function*::

    >>> t = numpy.linspace(-3, 3, 9)
    >>> print(numpy.around(distribution.pdf(t), 4))
    [0.0088 0.0209 0.0431 0.0775 0.121  0.1641 0.1933 0.1979 0.176 ]

Similarly to create values from the *cumulative distribution function*::

    >>> print(numpy.around(distribution.cdf(t), 4))
    [0.0062 0.0168 0.0401 0.0846 0.1587 0.266  0.4013 0.5497 0.6915]

To be able to perform any Monte Carlo method, each distribution contains
*random number generator*::

    >>> print(numpy.around(distribution.sample(6), 4))
    [ 2.7901 -0.4006  5.2952  1.9107  4.2763  0.4033]

The sample scheme also has a few advanced options. For example, to create
low-discrepancy Hammersley sequences samples combined with antithetic variates::

    >>> print(numpy.around(distribution.sample(
    ...     size=6, rule="H", antithetic=True), 4))
    [ 3.349   0.651  -0.3007  4.3007  2.6373  1.3627]

For a full overview of these options, see ..:ref:`sampling`

In addition a function for extracting raw-statistical moments is available::

    >>> print(distribution.mom([0, 1, 2, 3, 4]))
    [  1.   2.   8.  32. 160.]

Note that these are raw moments, not the classical moments with adjustments.
For example, the variance is defined as follows::

    >>> print(distribution.mom(2) - distribution.mom(1)**2)
    4.0

However, if the adjusted moments are of interest, the can be retrieved using
the tools described in :ref:`descriptives`::

    >>> print(chaospy.Var(distribution))
    4.0

Random Seed
-----------

To be able to reproduce results it is possible to fix the random seed in
``chaospy``. For simplicity, The library respect ``numpy.random.seed``. E.g.::

    >>> numpy.random.seed(1234)
    >>> print(numpy.around(distribution.sample(5), 4))
    [0.2554 2.622  1.6865 3.5808 3.5442]
    >>> numpy.random.seed(1234)
    >>> print(numpy.around(distribution.sample(5), 4))
    [0.2554 2.622  1.6865 3.5808 3.5442]
    >>> print(numpy.around(distribution.sample(5), 4))
    [0.79   0.8132 3.6967 5.459  4.3098]
"""
from . import baseclass
from .baseclass import Dist

from .sampler import *
from .operators import *
from .collection import *
from .copulas import *
from .evaluation import *
from .constructor import construct

from . import (
    sampler, evaluation, approximation,
    copulas, collection, operators, constructor
)



"""Generalized extreme value distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class generalized_extreme(Dist):
    """Generalized extreme value distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        cx = c*x
        logex2 = numpy.where(c == 0, 0., numpy.log1p(-cx))
        logpex2 = numpy.where(c == 0, -x, logex2/c)
        pex2 = numpy.exp(logpex2)
        logpdf = numpy.where((cx==1) | (cx==-numpy.inf), -numpy.inf, -pex2+logpex2-logex2)
        numpy.putmask(logpdf,(c==1) & (x==1),0.0)
        return numpy.exp(logpdf)

    def _cdf(self, x, c):
        loglogcdf = numpy.where(c == 0, -x, numpy.log1p(-c*x)/c)
        return numpy.exp(-numpy.exp(loglogcdf))

    def _ppf(self, q, c):
        x = -numpy.log(-numpy.log(q))
        return numpy.where(c == 0, x, -numpy.expm1(-c*x)/c)

    def _bnd(self, x, c):
        return self._ppf(1e-8, c), self._ppf(1-1e-8, c)


class GeneralizedExtreme(Add):
    """
    Generalized extreme value distribution
    Fisher-Tippett distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        loc (float, Dist) : Location parameter

    Example:
        >>> distribution = chaospy.GeneralizedExtreme(3, 2, 2)
        >>> print(distribution)
        GeneralizedExtreme(loc=2, scale=2, shape=3)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [-0.1126  2.1538  2.5778  2.6593]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0386 0.2382 1.1497 8.0333]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 2.6154 -4.0776  2.6666  2.4079]
        >>> print(numpy.around(distribution.mom(1), 4))
        -2.0561
    """

    def __init__(self, shape=0, scale=1, loc=0):
        self._repr = {"shape": shape, "scale": scale, "loc": loc}
        Add.__init__(self, left=generalized_extreme(shape)*scale, right=loc)


Genextreme = deprecation_warning(GeneralizedExtreme, "Genextreme")



"""
Newton-Cotes quadrature, are a group of formulas for numerical integration
based on evaluating the integrand at equally spaced points.

Example usage
-------------

Generate Newton-Cotes quadrature rules::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in range(5):
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="newton_cotes")
    ...     print(order, numpy.around(X, 3), numpy.around(W, 3))
    0 [[0.5]] [1.]
    1 [[0. 1.]] [0.5 0.5]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.    0.333 0.667 1.   ]] [0.125 0.375 0.375 0.125]
    4 [[0.   0.25 0.5  0.75 1.  ]] [0.078 0.356 0.133 0.356 0.078]

The first few orders with exponential growth rule::

    >>> for order in range(4):  # doctest: +NORMALIZE_WHITESPACE
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="newton_cotes", growth=True)
    ...     print(order, numpy.around(X, 3), numpy.around(W, 3))
    0 [[0.5]] [1.]
    1 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    2 [[0.   0.25 0.5  0.75 1.  ]] [0.078 0.356 0.133 0.356 0.078]
    3 [[0.    0.125 0.25  0.375 0.5   0.625 0.75  0.875 1.   ]]
       [ 0.035  0.208 -0.033  0.37  -0.16   0.37  -0.033  0.208  0.035]

Applying Smolyak sparse grid on Newton-Cotes::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> X, W = chaospy.generate_quadrature(
    ...     2, distribution, rule="newton_cotes",
    ...     growth=True, sparse=True)
    >>> print(numpy.around(X, 2))
    [[0.   0.   0.   0.25 0.5  0.5  0.5  0.5  0.5  0.75 1.   1.   1.  ]
     [0.   0.5  1.   0.5  0.   0.25 0.5  0.75 1.   0.5  0.   0.5  1.  ]]
    >>> print(numpy.around(W, 3))  # doctest: +NORMALIZE_WHITESPACE
    [ 0.028  0.022  0.028  0.356  0.022  0.356 -0.622
      0.356  0.022  0.356  0.028  0.022  0.028]
"""
from __future__ import print_function

import numpy
from scipy.integrate import newton_cotes

from .combine import combine_quadrature


def quad_newton_cotes(order, domain=(0, 1), growth=False):
    """
    Generate the abscissas and weights in Newton-Cotes quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Dist, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Examples:
        >>> abscissas, weights = quad_newton_cotes(3)
        >>> print(numpy.around(abscissas, 4))
        [[0.     0.3333 0.6667 1.    ]]
        >>> print(numpy.around(weights, 4))
        [0.375 1.125 1.125 0.375]
    """
    from ..distributions.baseclass import Dist
    if isinstance(domain, Dist):
        abscissas, weights = quad_newton_cotes(
            order, domain.range(), growth)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = domain
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()
    dim = max(lower.size, upper.size, order.size)
    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    results = [_newton_cotes(*args, growth=growth)
               for args in zip(order, lower, upper)]
    abscissas = [args[0] for args in results]
    weights = [args[1] for args in results]
    return combine_quadrature(abscissas, weights)


def _newton_cotes(order, lower, upper, growth):
    """Backend for Newton-Cotes quadrature rule."""
    if order == 0:
        return numpy.array([0.5*(lower+upper)]), numpy.ones(1)
    order = 2**order if growth else order
    return (
        numpy.linspace(lower, upper, order+1),
        newton_cotes(order)[0]/(upper-lower),
    )



from .alpha import Alpha
from .anglit import Anglit
from .beta import Beta, ArcSinus, Arcsinus, PowerLaw, Powerlaw, Wigner
from .binomial import Binomial
from .bradford import Bradford
from .burr import Burr
from .cauchy import Cauchy
from .chi import Chi, Maxwell, Rayleigh
from .chi_squared import ChiSquared, Chisquared
from .discrete_uniform import DiscreteUniform
from .double_gamma import DoubleGamma, Dbl_gamma
from .double_weibull import DoubleWeibull, Dbl_weibull
from .exponential_power import ExponentialPower, Exponpow
from .exponential_weibull import ExponentialWeibull, Exponweibull
from .f import F
from .fatigue_life import FatigueLife, Fatiguelife
from .fisk import Fisk
from .folded_cauchy import FoldedCauchy, Foldcauchy
from .folded_normal import FoldedNormal, Foldnormal
from .frechet import Frechet
from .gamma import Gamma, Exponential
from .generalized_exponential import GeneralizedExponential, Genexpon
from .generalized_extreme import GeneralizedExtreme, Genextreme
from .generalized_gamma import GeneralizedGamma, Gengamma
from .generalized_half_logistic import GeneralizedHalfLogistic, Genhalflogistic
from .gompertz import Gompertz
from .hyperbolic_secant import HyperbolicSecant, Hypgeosec
from .kumaraswamy import Kumaraswamy
from .laplace import Laplace
from .levy import Levy
from .log_gamma import LogGamma, Loggamma
from .log_laplace import LogLaplace, Loglaplace
from .log_normal import LogNormal, Lognormal, Gilbrat
from .log_uniform import LogUniform, Loguniform
from .log_weibull import LogWeibull, Logweibull
from .logistic import Logistic
from .mielke import Mielke
from .mv_log_normal import MvLogNormal, MvLognormal
from .mv_normal import MvNormal
from .mv_student_t import MvStudentT, MvStudent_t
from .nakagami import Nakagami
from .normal import Normal
from .pareto1 import Pareto1
from .pareto2 import Pareto2
from .power_log_normal import PowerLogNormal, PowerLognormal
from .power_normal import PowerNormal, Powernorm
from .reciprocal import Reciprocal
from .student_t import StudentT, Student_t
from .triangle import Triangle
from .trunc_exponential import TruncExponential, Truncexpon
from .trunc_normal import TruncNormal, Truncnorm
from .tukey_lambda import TukeyLambda, Tukeylambda
from .uniform import Uniform
from .wald import Wald
from .weibull import Weibull
from .wrapped_cauchy import WrappedCauchy



"""Variance operator."""
import numpy

from .. import distributions, poly as polynomials


def Var(poly, dist=None, **kws):
    """
    Element by element 2nd order statistics.

    Args:
        poly (Poly, Dist):
            Input to take variance on.
        dist (Dist):
            Defines the space the variance is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``variation.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> print(chaospy.Var(dist))
        [1. 4.]
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> print(chaospy.Var(poly, dist))
        [  0.   1.   4. 800.]
    """
    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    dim = len(dist)
    if poly.dim<dim:
        polynomials.setdim(poly, dim)

    shape = poly.shape
    poly = polynomials.flatten(poly)

    keys = poly.keys
    N = len(keys)
    A = poly.A

    keys1 = numpy.array(keys).T
    if dim==1:
        keys1 = keys1[0]
        keys2 = sum(numpy.meshgrid(keys, keys))
    else:
        keys2 = numpy.empty((dim, N, N))
        for i in range(N):
            for j in range(N):
                keys2[:, i, j] = keys1[:, i]+keys1[:, j]

    m1 = numpy.outer(*[dist.mom(keys1, **kws)]*2)
    m2 = dist.mom(keys2, **kws)
    mom = m2-m1

    out = numpy.zeros(poly.shape)
    for i in range(N):
        a = A[keys[i]]
        out += a*a*mom[i, i]
        for j in range(i+1, N):
            b = A[keys[j]]
            out += 2*a*b*mom[i, j]

    out = out.reshape(shape)
    return out



"""Gompertz distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class gompertz(Dist):
    """Gompertz distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        ex = numpy.exp(x)
        return c*ex*numpy.exp(-c*(ex-1))

    def _cdf(self, x, c):
        return 1.0-numpy.exp(-c*(numpy.exp(x)-1))

    def _ppf(self, q, c):
        return numpy.log(1-1.0/c*numpy.log(1-q))

    def _bnd(self, x, c):
        return 0., self._ppf(1-1e-10, c)


class Gompertz(Add):
    """
    Gompertz distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Gompertz(3, 2, 2)
        >>> print(distribution)
        Gompertz(scale=2, shape=3, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.1435 2.3145 2.5331 2.859 ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.2893 1.0532 0.7833 0.4609]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.6052 2.0798 3.3868 2.3967]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.5242
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.034  3.3243 3.5329]
         [0.1878 0.3655 0.5218]]
    """

    def __init__(self, shape, scale, shift):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=gompertz(shape)*scale, right=shift)



"""Beta distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class beta_(Dist):

    def __init__(self, a=1, b=1):
        Dist.__init__(self, a=a, b=b)

    def _pdf(self, x, a, b):
        return x**(a-1)*(1-x)**(b-1)/ \
            special.beta(a, b)

    def _cdf(self, x, a, b):
        return special.btdtr(a, b, x)

    def _ppf(self, q, a, b):
        return special.btdtri(a, b, q)

    def _mom(self, k, a, b):
        return special.beta(a+k,b)/special.beta(a,b)

    def _ttr(self, n, a, b):
        nab = 2*n+a+b
        A = ((a-1)**2-(b-1)**2)*.5/\
                (nab*(nab-2) + (nab==0) + (nab==2)) + .5
        B1 = a*b*1./((a+b+1)*(a+b)**2)
        B2 = (n+a-1)*(n+b-1)*n*(n+a+b-2.)/\
            ((nab-1)*(nab-3)*(nab-2)**2+2.*((n==0)+(n==1)))
        B = numpy.where((n==0)+(n==1), B1, B2)
        return A, B

    def _bnd(self, x, a, b):
        return 0., 1.


class Beta(Add):
    R"""
    Beta Probability Distribution.

    Args:
        alpha (float, Dist): First shape parameter, alpha > 0
        beta (float, Dist): Second shape parameter, b > 0
        lower (float, Dist): Lower threshold
        upper (float, Dist): Upper threshold

    Examples:
        >>> distribution = chaospy.Beta(2, 2, 2, 3)
        >>> print(distribution)
        Beta(alpha=2, beta=2, lower=2, upper=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.2871 2.4329 2.5671 2.7129]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.2281 1.473  1.473  1.2281]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.6039 2.2112 2.8651 2.4881]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.5
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[2.5    2.5    2.5   ]
         [0.05   0.0571 0.0595]]
    """

    def __init__(self, alpha, beta, lower=0, upper=1):
        self._repr = {
            "alpha": alpha, "beta": beta, "lower": lower, "upper": upper}
        Add.__init__(self, left=beta_(alpha, beta)*(upper-lower), right=lower)


class ArcSinus(Add):
    """
    Generalized Arc-sinus distribution

    Args:
        shape (float, Dist): Shape parameter where 0.5 is the default
            non-generalized case. Defined on the interval ``[0, 1]``.
        lower (float, Dist): Lower threshold
        upper (float, Dist): Upper threshold

    Examples:
        >>> distribution = chaospy.ArcSinus(0.8, 4, 6)
        >>> print(distribution)
        ArcSinus(lower=4, shape=0.8, upper=6)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [4.9875 5.6438 5.9134 5.9885 5.9996]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [ 0.1857  0.3868  1.1633  5.8145 92.8592]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.9861 4.6882 6.     5.897 ]
        >>> print(distribution.mom(1))
        5.6
    """

    def __init__(self, shape=0.5, lower=0, upper=1):
        self._repr = {"shape": shape, "lower": lower, "upper": upper}
        Add.__init__(
            self, left=beta_(shape, 1-shape)*(upper-lower), right=lower)

Arcsinus = deprecation_warning(ArcSinus, "Arcsinus")


class PowerLaw(Add):
    """
    Powerlaw distribution

    Args:
        shape (float, Dist) : Shape parameter
        lower (float, Dist) : Location of lower threshold
        upper (float, Dist) : Location of upper threshold

    Examples:
        >>> distribution = chaospy.PowerLaw(0.8, 4, 6)
        >>> print(distribution)
        PowerLaw(lower=4, shape=0.8, upper=6)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [4.213  4.5066 4.8409 5.2048 5.5924]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.626  0.5264 0.4757 0.4427 0.4187]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.1753 4.1339 5.8765 4.8036]
        >>> print(numpy.around(distribution.mom(1), 4))
        4.8889
    """

    def __init__(self, shape=1, lower=0, upper=1):
        self._repr = {"shape": shape, "lower": lower, "upper": upper}
        Add.__init__(self, left=beta_(shape, 1)*(upper-lower), right=lower)


Powerlaw = deprecation_warning(PowerLaw, "Powerlaw")


class Wigner(Add):
    """
    Wigner (semi-circle) distribution

    Args:
        radius (float, Dist) : radius of the semi-circle (scale)
        shift (float, Dist) : location of the circle origin (location)

    Examples:
        >>> distribution = chaospy.Wigner(2, 3)
        >>> print(distribution)
        Wigner(radius=2, shift=3)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.8934 2.4701 3.     3.5299 4.1066]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.2651 0.3069 0.3183 0.3069 0.2651]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.4874 1.6895 4.6123 2.944 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.0
    """

    def __init__(self, radius=1, shift=0):
        self._repr = {"radius": radius, "shift": shift}
        Add.__init__(self, left=radius*(2*beta_(1.5, 1.5)-1), right=shift)



"""Log-uniform distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class log_uniform(Dist):
    """Log-uniform distribution."""

    def __init__(self, lo=0, up=1):
        Dist.__init__(self, lo=lo, up=up)

    def _pdf(self, x, lo, up):
        return 1./(x*(up-lo))

    def _cdf(self, x, lo, up):
        return (numpy.log(x)-lo)/(up-lo)

    def _ppf(self, q, lo, up):
        return numpy.e**(q*(up-lo) + lo)

    def _bnd(self, x, lo, up):
        return numpy.e**lo, numpy.e**up

    def _mom(self, k, lo, up):
        return ((numpy.e**(up*k)-numpy.e**(lo*k))/((up-lo)*(k+(k==0))))**(k!=0)


class LogUniform(Add):
    """
    Log-uniform distribution

    Args:
        lower (float, Dist): Location of lower threshold of uniform distribution.
        upper (float, Dist): Location of upper threshold of uniform distribution.
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.LogUniform(2, 3, 2, 3)
        >>> print(distribution)
        LogUniform(lower=2, scale=2, shift=3, upper=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [21.05   25.0464 29.9275 35.8893]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0554 0.0454 0.0371 0.0304]
        >>> print(numpy.around(distribution.sample(4), 4))
        [31.4099 19.5793 41.2227 26.9349]
        >>> print(numpy.around(distribution.mom(1), 4))
        28.393
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[30.8948 30.5352 30.4949]
         [52.8588 42.8862 41.419 ]]
    """

    def __init__(self, lower=0, upper=1, scale=1, shift=0):
        self._repr = {"lower": lower, "upper": upper, "scale": scale, "shift": shift}
        Add.__init__(self, left=log_uniform(lower, upper)*scale, right=shift)


Loguniform = deprecation_warning(LogUniform, "Loguniform")



"""Frechet or Extreme value distribution type 2."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class frechet(Dist):
    """Frechet or Extreme value distribution type 2."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return c*pow(x,c-1)*numpy.exp(-pow(x,c))

    def _cdf(self, x, c):
        return -numpy.expm1(-pow(x,c))

    def _ppf(self, q, c):
        return pow(-numpy.log1p(-q),1.0/c)

    def _mom(self, k, c):
        return special.gamma(1-k*1./c)

    def _bnd(self, x, c):
        return 0, self._ppf(1-1e-10, c)


class Frechet(Add):
    """
    Frechet or Extreme value distribution type 2.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.Frechet(3, 2, 1)
        >>> print(distribution)
        Frechet(scale=2, shape=3, shift=1)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.2131 2.5988 2.9426 3.3438]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.4415 0.5751 0.566  0.412 ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.0393 1.9924 3.8849 2.7397]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.7082
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[2.8951 3.039  3.1808]
         [0.4213 0.7168 0.9441]]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=frechet(shape)*scale, right=shift)



"""
Evaluate inverse Rosenblatt transformation.

In the case of univariate distributions, this is equivalent to the point
percentile function (PPF).

Example usage
-------------

Define a simple distribution and data::

    >>> class Exponential(chaospy.Dist):
    ...     def _ppf(self, u_data, alpha): return -numpy.log(1-u_data)/alpha
    >>> dist = Exponential(alpha=2)
    >>> u_data = numpy.array([[0.1, 0.2, 0.3]])

Normal usage::

    >>> print(numpy.around(evaluate_inverse(dist, u_data), 4))
    [[0.0527 0.1116 0.1783]]

Use non-default parameters::

    >>> print(numpy.around(evaluate_inverse(
    ...     dist, u_data, parameters={"alpha": 1}), 4))
    [[0.1054 0.2231 0.3567]]

If a distribution is missing the definition of the density function, it is
instead estimated from cumulative distribution function and boundary function::

    >>> class Exponential(chaospy.Dist):
    ...     def _cdf(self, x_data, alpha): return 1-numpy.e**(-alpha*x_data)
    ...     def _bnd(self, x_data, alpha): return 0, 100
    >>> dist = Exponential(alpha=1)
    >>> print(numpy.around(evaluate_inverse(dist, u_data), 4))
    [[0.1054 0.2231 0.3567]]
"""
import numpy
from .parameters import load_parameters


def evaluate_inverse(
        distribution,
        u_data,
        cache=None,
        parameters=None
):
    """
    Evaluate inverse Rosenblatt transformation.

    Args:
        distribution (Dist):
            Distribution to evaluate.
        u_data (numpy.ndarray):
            Locations for where evaluate inverse transformation distribution at.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The cumulative distribution values of ``distribution`` at location
        ``u_data`` using parameters ``parameters``.
    """
    if cache is None:
        cache = {}
    out = numpy.zeros(u_data.shape)

    # Distribution self know how to handle inverse Rosenblatt.
    if hasattr(distribution, "_ppf"):
        parameters = load_parameters(
            distribution, "_ppf", parameters=parameters, cache=cache)
        out[:] = distribution._ppf(u_data.copy(), **parameters)

    # Approximate inverse Rosenblatt based on cumulative distribution function.
    else:
        from .. import approximation
        parameters = load_parameters(
            distribution, "_cdf", parameters=parameters, cache=cache)
        out[:] = approximation.approximate_inverse(
            distribution, u_data.copy(), cache=cache.copy(), parameters=parameters)

    # Store cache.
    cache[distribution] = out

    return out



"""Logarithm with base N."""
import numpy

from .. import evaluation, approximation
from ..baseclass import Dist


class Logn(Dist):
    """
    Logarithm with base N.

    Args:
        dist (Dist): Distribution to perform transformation on.
        base (int, float): the logarithm base.

    Example:
        >>> distribution = chaospy.Logn(chaospy.Uniform(1, 2), 3)
        >>> print(distribution)
        Logn(Uniform(lower=1, upper=2), 3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.166  0.3063 0.4278 0.535 ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.3183 1.5381 1.7578 1.9775]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.4578 0.0991 0.608  0.3582]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.3516
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.3516 0.3085 0.3144]
         [1.     0.0324 0.0266]]
    """

    def __init__(self, dist, base=2):
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range() > 0)
        assert base > 0 and base != 1
        Dist.__init__(self, dist=dist, base=base)

    def _pdf(self, xloc, dist, base, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, base**xloc, cache=cache)*base**xloc*numpy.log(base)

    def _cdf(self, xloc, dist, base, cache):
        """Cumulative distribution function."""
        return evaluation.evaluate_forward(dist, base**xloc, cache=cache)

    def _ppf(self, q, dist, base, cache):
        """Point percentile function."""
        return numpy.log(evaluation.evaluate_inverse(
            dist, q, cache=cache))/numpy.log(base)

    def _bnd(self, xloc, dist, base, cache):
        """Distribution bounds."""
        return numpy.log(evaluation.evaluate_bound(
            dist, base**xloc, cache=cache)) / numpy.log(base)

    def _mom(self, x, dist, base, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return (self.__class__.__name__ + "(" + str(self.prm["dist"]) + ", " +
                str(self.prm["base"]) + ")")

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        base = evaluation.get_forward_cache(self.prm["base"], cache)
        if not isinstance(dist, Dist) and not isinstance(base, Dist):
            return numpy.log(dist) / numpy.log(base)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        base = evaluation.get_forward_cache(self.prm["base"], cache)
        if not isinstance(dist, Dist) and not isinstance(base, Dist):
            return base**dist
        return self



"""Cauchy distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class cauchy(Dist):
    """Standard Cauchy distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return 1.0/numpy.pi/(1.0+x*x)

    def _cdf(self, x):
        return 0.5 + 1.0/numpy.pi*numpy.arctan(x)

    def _ppf(self, q):
        return numpy.tan(numpy.pi*q-numpy.pi/2.0)

    def _bnd(self, x):
        return self._ppf(1e-10), self._ppf(1-1e-10)


class Cauchy(Add):
    """
    Cauchy distribution.

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter

    Examples:
        >>> distribution = chaospy.Cauchy(2, 4)
        >>> print(distribution)
        Cauchy(loc=2, scale=4)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [-4.9282 -0.3094  2.      4.3094  8.9282]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0199 0.0597 0.0796 0.0597 0.0199]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 4.0953 -8.585  27.4011  1.776 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
    """

    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=cauchy()*scale, right=loc)



import numpy as np

import chaospy.poly
from chaospy.poly.base import Poly


def call(poly, args):
    """
    Evaluate a polynomial along specified axes.

    Args:
        poly (Poly):
            Input polynomial.
        args (numpy.ndarray):
            Argument to be evaluated. Masked values keeps the variable intact.

    Returns:
        (Poly, numpy.ndarray):
            If masked values are used the Poly is returned. Else an numpy array
            matching the polynomial's shape is returned.
    """
    args = list(args)

    # expand args to match dim
    if len(args) < poly.dim:
        args = args + [np.nan]*(poly.dim-len(args))

    elif len(args) > poly.dim:
        raise ValueError("too many arguments")

    # Find and perform substitutions, if any
    x0, x1 = [], []
    for idx, arg in enumerate(args):

        if isinstance(arg, Poly):
            poly_ = Poly({
                tuple(np.eye(poly.dim)[idx]): np.array(1)
            })
            x0.append(poly_)
            x1.append(arg)
            args[idx] = np.nan
    if x0:
        poly = call(poly, args)
        return substitute(poly, x0, x1)

    # Create masks
    masks = np.zeros(len(args), dtype=bool)
    for idx, arg in enumerate(args):
        if np.ma.is_masked(arg) or np.any(np.isnan(arg)):
            masks[idx] = True
            args[idx] = 0

    shape = np.array(
        args[
            np.argmax(
                [np.prod(np.array(arg).shape) for arg in args]
            )
        ]
    ).shape
    args = np.array([np.ones(shape, dtype=int)*arg for arg in args])

    A = {}
    for key in poly.keys:

        key_ = np.array(key)*(1-masks)
        val = np.outer(poly.A[key], np.prod((args.T**key_).T, \
                axis=0))
        val = np.reshape(val, poly.shape + tuple(shape))
        val = np.where(val != val, 0, val)

        mkey = tuple(np.array(key)*(masks))
        if not mkey in A:
            A[mkey] = val
        else:
            A[mkey] = A[mkey] + val

    out = Poly(A, poly.dim, None, None)
    if out.keys and not np.sum(out.keys):
        out = out.A[out.keys[0]]
    elif not out.keys:
        out = np.zeros(out.shape, dtype=out.dtype)
    return out


def substitute(P, x0, x1, V=0):
    """
    Substitute a variable in a polynomial array.

    Args:
        P (Poly) : Input data.
        x0 (Poly, int) : The variable to substitute. Indicated with either unit
                variable, e.g. `x`, `y`, `z`, etc. or through an integer
                matching the unit variables dimension, e.g. `x==0`, `y==1`,
                `z==2`, etc.
        x1 (Poly) : Simple polynomial to substitute `x0` in `P`. If `x1` is an
                polynomial array, an error will be raised.

    Returns:
        (Poly) : The resulting polynomial (array) where `x0` is replaced with
                `x1`.

    Examples:
        >>> x,y = cp.variable(2)
        >>> P = cp.Poly([y*y-1, y*x])
        >>> print(cp.substitute(P, y, x+1))
        [q0^2+2q0, q0^2+q0]

        With multiple substitutions:
        >>> print(cp.substitute(P, [x,y], [y,x]))
        [q0^2-1, q0q1]
    """
    x0,x1 = map(Poly, [x0,x1])
    dim = np.max([p.dim for p in [P,x0,x1]])
    dtype = chaospy.poly.typing.dtyping(P.dtype, x0.dtype, x1.dtype)
    P, x0, x1 = [chaospy.poly.dimension.setdim(p, dim) for p in [P,x0,x1]]

    if x0.shape:
        x0 = [x for x in x0]
    else:
        x0 = [x0]

    if x1.shape:
        x1 = [x for x in x1]
    else:
        x1 = [x1]

    # Check if substitution is needed.
    valid = False
    C = [x.keys[0].index(1) for x in x0]
    for key in P.keys:
        if np.any([key[c] for c in C]):
            valid = True
            break

    if not valid:
        return P

    dims = [tuple(np.array(x.keys[0])!=0).index(True) for x in x0]

    dec = is_decomposed(P)
    if not dec:
        P = decompose(P)

    P = chaospy.poly.dimension.dimsplit(P)

    shape = P.shape
    P = [p for p in chaospy.poly.shaping.flatten(P)]

    for i in range(len(P)):
        for j in range(len(dims)):
            if P[i].keys and P[i].keys[0][dims[j]]:
                P[i] = x1[j].__pow__(P[i].keys[0][dims[j]])
                break

    P = Poly(P, dim, None, dtype)
    P = chaospy.poly.shaping.reshape(P, shape)
    P = chaospy.poly.collection.prod(P, 0)

    if not dec:
        P = chaospy.poly.collection.sum(P, 0)

    return P


def is_decomposed(P):
    """
    Check if a polynomial (array) is on component form.

    Args:
        P (Poly):
            Input data.

    Returns:
        (bool):
            True if all polynomials in ``P`` are on component form.

    Examples:
        >>> x,y = cp.variable(2)
        >>> print(cp.is_decomposed(cp.Poly([1,x,x*y])))
        True
        >>> print(cp.is_decomposed(cp.Poly([x+1,x*y])))
        False
    """
    if P.shape:
        return min([is_decomposed(poly) for poly in P])
    return len(P.keys) <= 1


def decompose(P):
    """
    Decompose a polynomial to component form.

    In array missing values are padded with 0 to make decomposition compatible
    with ``chaospy.sum(Q, 0)``.

    Args:
        P (Poly) : Input data.

    Returns:
        (Poly) : Decomposed polynomial with `P.shape==(M,)+Q.shape` where
                `M` is the number of components in `P`.

    Examples:
        >>> q = cp.variable()
        >>> P = cp.Poly([q**2-1, 2])
        >>> print(P)
        [q0^2-1, 2]
        >>> print(cp.decompose(P))
        [[-1, 2], [q0^2, 0]]
        >>> print(cp.sum(cp.decompose(P), 0))
        [q0^2-1, 2]
    """
    P = P.copy()

    if not P:
        return P

    out = [Poly({key:P.A[key]}) for key in P.keys]
    return Poly(out, None, None, None)


import chaospy as cp



import numpy

from ..conditional import E_cond
from ..expected import E
from ..variance import Var


def Sens_m2(poly, dist, **kws):
    """
    Variance-based decomposition/Sobol' indices.

    Second order sensitivity indices.

    Args:
        poly (Poly):
            Polynomial to find second order Sobol indices on.
        dist (Dist):
            The distributions of the input used in ``poly``.

    Returns:
        (numpy.ndarray):
            First order sensitivity indices for each parameters in ``poly``,
            with shape ``(len(dist), len(dist)) + poly.shape``.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x*y, x*x*y*y, x*y*y*y])
        >>> dist = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> indices = chaospy.Sens_m2(poly, dist)
        >>> print(indices)
        [[[0.         0.         0.         0.        ]
          [0.         0.14285714 0.28571429 0.20930233]]
        <BLANKLINE>
         [[0.         0.14285714 0.28571429 0.20930233]
          [0.         0.         0.         0.        ]]]
    """
    dim = len(dist)
    if poly.dim<dim:
        poly = chaospy.poly.setdim(poly, len(dist))

    zero = [0]*dim
    out = numpy.zeros((dim, dim) + poly.shape)

    mean = E(poly, dist)
    V_total = Var(poly, dist)
    E_cond_i = [None]*dim
    V_E_cond_i = [None]*dim
    for i in range(dim):
        zero[i] = 1
        E_cond_i[i] = E_cond(poly, zero, dist, **kws) 
        V_E_cond_i[i] = Var(E_cond_i[i], dist, **kws)
        zero[i] = 0

    for i in range(dim):

        zero[i] = 1
        for j in range(i+1, dim):

            zero[j] = 1
            E_cond_ij = E_cond(poly, zero, dist, **kws)
            out[i, j] = ((Var(E_cond_ij, dist, **kws)-V_E_cond_i[i] - V_E_cond_i[j]) /
                         (V_total+(V_total == 0))*(V_total != 0))
            out[j, i] = out[i, j]
            zero[j] = 0

        zero[i] = 0

    return out



"""Kurtosis operator."""
import numpy

from .. import distributions, poly as polynomials
from .expected import E


def Kurt(poly, dist=None, fisher=True, **kws):
    """
    Kurtosis operator.

    Element by element 4rd order statistics of a distribution or polynomial.

    Args:
        poly (Poly, Dist):
            Input to take kurtosis on.
        dist (Dist):
            Defines the space the skewness is taken on. It is ignored if
            ``poly`` is a distribution.
        fisher (bool):
            If True, Fisher's definition is used (Normal -> 0.0). If False,
            Pearson's definition is used (normal -> 3.0)

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``skewness.shape==poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> print(numpy.around(chaospy.Kurt(dist), 4))
        [6. 0.]
        >>> print(numpy.around(chaospy.Kurt(dist, fisher=False), 4))
        [9. 3.]
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> print(numpy.around(chaospy.Kurt(poly, dist), 4))
        [nan  6.  0. 15.]
    """
    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    if fisher:
        adjust = 3
    else:
        adjust = 0

    shape = poly.shape
    poly = polynomials.flatten(poly)

    m1 = E(poly, dist)
    m2 = E(poly**2, dist)
    m3 = E(poly**3, dist)
    m4 = E(poly**4, dist)

    out = (m4-4*m3*m1 + 6*m2*m1**2 - 3*m1**4) /\
            (m2**2-2*m2*m1**2+m1**4) - adjust

    out = numpy.reshape(out, shape)
    return out



"""
Create `Van Der Corput` low discrepancy sequence samples.

A van der Corput sequence is an example of the simplest one-dimensional
low-discrepancy sequence over the unit interval; it was first described in 1935
by the Dutch mathematician J. G. van der Corput. It is constructed by reversing
the base-n representation of the sequence of natural numbers (1, 2, 3, ...).

In practice, use Halton sequence instead of Van Der Corput, as it is the
same, but generalized to work in multiple dimensions.

Example usage
-------------

Using base 10::

    >>> print(create_van_der_corput_samples(range(11), number_base=10))
    [0.1  0.2  0.3  0.4  0.5  0.6  0.7  0.8  0.9  0.01 0.11]

Using base 2::

    >>> print(create_van_der_corput_samples(range(8), number_base=2))
    [0.5    0.25   0.75   0.125  0.625  0.375  0.875  0.0625]

.. Van Der Corput: https://en.wikipedia.org/wiki/Van_der_Corput_sequence
"""
from __future__ import division
import numpy


def create_van_der_corput_samples(idx, number_base=2):
    """
    Van der Corput samples.

    Args:
        idx (int, numpy.ndarray):
            The index of the sequence. If array is provided, all values in
            array is returned.
        number_base (int):
            The numerical base from where to create the samples from.

    Returns (float, numpy.ndarray):
        Van der Corput samples.
    """
    assert number_base > 1

    idx = numpy.asarray(idx).flatten() + 1
    out = numpy.zeros(len(idx), dtype=float)

    base = float(number_base)
    active = numpy.ones(len(idx), dtype=bool)
    while numpy.any(active):
        out[active] += (idx[active] % number_base)/base
        idx //= number_base
        base *= number_base
        active = idx > 0
    return out



"""
Function that overlaps with the numpy library.
"""
import numpy as np

import chaospy as cp
import chaospy.poly
from chaospy.poly.base import Poly


def sum(vari, axis=None): # pylint: disable=redefined-builtin
    """
    Sum the components of a shapeable quantity along a given axis.

    Args:
        vari (chaospy.poly.base.Poly, numpy.ndarray):
            Input data.
        axis (int):
            Axis over which the sum is taken. By default ``axis`` is None, and
            all elements are summed.

    Returns:
        (chaospy.poly.base.Poly, numpy.ndarray):
            Polynomial array with same shape as ``vari``, with the specified
            axis removed. If ``vari`` is an 0-d array, or ``axis`` is None,
            a (non-iterable) component is returned.

    Examples:
        >>> vari = cp.prange(3)
        >>> print(vari)
        [1, q0, q0^2]
        >>> print(cp.sum(vari))
        q0^2+q0+1
    """
    if isinstance(vari, Poly):

        core = vari.A.copy()
        for key in vari.keys:
            core[key] = sum(core[key], axis)

        return Poly(core, vari.dim, None, vari.dtype)

    return np.sum(vari, axis)


def cumsum(vari, axis=None):
    """
    Cumulative sum the components of a shapeable quantity along a given axis.

    Args:
        vari (chaospy.poly.base.Poly, numpy.ndarray):
            Input data.
        axis (int):
            Axis over which the sum is taken. By default ``axis`` is None, and
            all elements are summed.

    Returns:
        (chaospy.poly.base.Poly, numpy.ndarray):
            Polynomial array with same shape as ``vari``.

    Examples:
        >>> poly = cp.prange(3)
        >>> print(poly)
        [1, q0, q0^2]
        >>> print(cp.cumsum(poly))
        [1, q0+1, q0^2+q0+1]
    """
    if isinstance(vari, Poly):
        core = vari.A.copy()
        for key, val in core.items():
            core[key] = cumsum(val, axis)
        return Poly(core, vari.dim, None, vari.dtype)

    return np.cumsum(vari, axis)


def prod(vari, axis=None):
    """
    Product of the components of a shapeable quantity along a given axis.

    Args:
        vari (chaospy.poly.base.Poly, numpy.ndarray):
            Input data.
        axis (int):
            Axis over which the sum is taken. By default ``axis`` is None, and
            all elements are summed.

    Returns:
        (chaospy.poly.base.Poly, numpy.ndarray):
            Polynomial array with same shape as ``vari``, with the specified
            axis removed. If ``vari`` is an 0-d array, or ``axis`` is None,
            a (non-iterable) component is returned.

    Examples:
        >>> vari = cp.prange(3)
        >>> print(vari)
        [1, q0, q0^2]
        >>> print(cp.prod(vari))
        q0^3
    """
    if isinstance(vari, Poly):
        if axis is None:
            vari = chaospy.poly.shaping.flatten(vari)
            axis = 0

        vari = chaospy.poly.shaping.rollaxis(vari, axis)
        out = vari[0]
        for poly in vari[1:]:
            out = out*poly
        return out

    return np.prod(vari, axis)


def cumprod(vari, axis=None):
    """
    Perform the cumulative product of a shapeable quantity over a given axis.

    Args:
        vari (chaospy.poly.base.Poly, numpy.ndarray):
            Input data.
        axis (int):
            Axis over which the sum is taken. By default ``axis`` is None, and
            all elements are summed.

    Returns:
        (chaospy.poly.base.Poly):
            An array shaped as ``vari`` but with the specified axis removed.

    Examples:
        >>> vari = cp.prange(4)
        >>> print(vari)
        [1, q0, q0^2, q0^3]
        >>> print(cp.cumprod(vari))
        [1, q0, q0^3, q0^6]
    """
    if isinstance(vari, Poly):
        if np.prod(vari.shape) == 1:
            return vari.copy()
        if axis is None:
            vari = chaospy.poly.shaping.flatten(vari)
            axis = 0

        vari = chaospy.poly.shaping.rollaxis(vari, axis)
        out = [vari[0]]

        for poly in vari[1:]:
            out.append(out[-1]*poly)
        return Poly(out, vari.dim, vari.shape, vari.dtype)

    return np.cumprod(vari, axis)



"""Gumbel or Log-Weibull distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class log_weibull(Dist):
    """Gumbel or Log-Weibull distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        ex = numpy.exp(-x)
        return ex*numpy.exp(-ex)

    def _cdf(self, x):
        return numpy.exp(-numpy.exp(-x))

    def _ppf(self, q):
        return -numpy.log(-numpy.log(q))

    def _bnd(self, x):
        return self._ppf(1e-10), self._ppf(1-1e-10)


class LogWeibull(Add):
    """
    Gumbel or Log-Weibull distribution.

    Args:
        scale (float, Dist) : Scaling parameter
        loc (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.LogWeibull(2, 2)
        >>> print(distribution)
        LogWeibull(loc=2, scale=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.0482 2.1748 3.3435 4.9999]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1609 0.1833 0.1532 0.0893]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.71   0.4572 7.952  2.631 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.1544
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 6.0775  9.5552 13.213 ]
         [ 6.5797 20.4065 42.3767]]
    """
    def __init__(self, scale=1, loc=0):
        self._repr = {"scale": scale, "loc": loc}
        Add.__init__(self, left=log_weibull()*scale, right=loc)


Logweibull = deprecation_warning(LogWeibull, "Logweibull")



"""Generalized half-logistic distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class generalized_half_logistic(Dist):
    """Generalized half-logistic distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        limit = 1.0/c
        tmp = (1-c*x)
        tmp0 = tmp**(limit-1)
        tmp2 = tmp0*tmp
        return 2*tmp0 / (1+tmp2)**2

    def _cdf(self, x, c):
        limit = 1.0/c
        tmp = (1-c*x)
        tmp2 = tmp**(limit)
        return (1.0-tmp2) / (1+tmp2)

    def _ppf(self, q, c):
        return 1.0/c*(1-((1.0-q)/(1.0+q))**c)

    def _bnd(self, x, c):
        return 0.0, 1/numpy.where(c<10**-10, 10**-10, c)


class GeneralizedHalfLogistic(Add):
    """
    Generalized half-logistic distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.GeneralizedHalfLogistic(1, 2, 2)
        >>> print(distribution)
        GeneralizedHalfLogistic(scale=2, shape=1, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.6667 3.1429 3.5    3.7778]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.36 0.49 0.64 0.81]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.581  2.4126 3.949  3.3013]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.2274
    """

    def __init__(self, shape, scale, shift):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=generalized_half_logistic(shape)*scale, right=shift)



Genhalflogistic = deprecation_warning(GeneralizedHalfLogistic, "Genhalflogistic")



"""Nataf (normal) copula."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist

class nataf(Dist):
    """Nataf (normal) copula."""

    def __init__(self, R, ordering=None):
        if ordering is None:
            ordering = range(len(R))
        ordering = numpy.array(ordering)
        P = numpy.eye(len(R))[ordering]
        R = numpy.dot(P, numpy.dot(R, P.T))
        R = numpy.linalg.cholesky(R)
        R = numpy.dot(P.T, numpy.dot(R, P))
        Ci = numpy.linalg.inv(R)
        Dist.__init__(self, C=R, Ci=Ci)
        self.length = len(R)

    def __len__(self):
        return self.length

    def _cdf(self, x, C, Ci):
        out = special.ndtr(numpy.dot(Ci, special.ndtri(x)))
        return out

    def _ppf(self, q, C, Ci):
        out = special.ndtr(numpy.dot(C, special.ndtri(q)))
        return out

    def _bnd(self, C, Ci):
        return 0., 1.

class Nataf(Copula):
    """
    Nataf (normal) copula.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Uniform(), 2)
        >>> R = [[1, 0.5], [0.5, 1]]
        >>> copula = chaospy.Nataf(distribution, R)
        >>> print(copula)
        Nataf(Iid(Uniform(lower=0, upper=1), 2), R=[[1, 0.5], [0.5, 1]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(copula.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.1784 0.2796 0.4025]
          [0.368  0.5    0.632 ]
          [0.5975 0.7204 0.8216]]]
        >>> print(numpy.around(copula.fwd(copula.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(copula.pdf(copula.inv(mesh)), 4))
        [[1.4061 1.0909 0.9482]
         [1.2223 1.1547 1.2223]
         [0.9482 1.0909 1.4061]]
        >>> print(numpy.around(copula.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.8816 0.0983 0.2466 0.4021]]
    """

    def __init__(self, dist, R, ordering=None):
        """
        Args:
            dist (Dist) : The Distribution to wrap.
            R (numpy.ndarray) : Covariance matrix.
        """
        self._repr = {"R": R}
        assert len(dist) == len(R)
        return Copula.__init__(self, dist=dist, trans=nataf(R, ordering))



"""Reciprocal distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class Reciprocal(Dist):
    """
    Reciprocal distribution.

    Args:
        lower (float, Dist): Lower threshold of distribution. Must be smaller
            than ``upper``.
        upper (float, Dist): Upper threshold of distribution.

    Examples:
        >>> distribution = chaospy.Reciprocal(2, 4)
        >>> print(distribution)
        Reciprocal(lower=2, upper=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.     2.3784 2.8284 3.3636 4.    ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.7213 0.6066 0.5101 0.4289 0.3607]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.1462 2.166  3.8645 2.7937]
        >>> print(numpy.around(distribution.mom(1), 4))
        7.8433
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.023  3.0033 3.0011]
         [0.3307 0.2664 0.257 ]]
    """

    def __init__(self, lower=1, upper=2):
        self._repr = {"lower": lower, "upper": upper}
        Dist.__init__(self, lower=lower, upper=upper)

    def _pdf(self, x, lower, upper):
        return 1./(x*numpy.log(upper/lower))

    def _cdf(self, x, lower, upper):
        return numpy.log(x/lower)/numpy.log(upper/lower)

    def _ppf(self, q, lower, upper):
        return numpy.e**(q*numpy.log(upper/lower) + numpy.log(lower))

    def _bnd(self, x, lower, upper):
        return lower, upper

    def _mom(self, k, lower, upper):
        return ((upper*numpy.e**k-lower*numpy.e**k)/(numpy.log(upper/lower)*(k+(k == 0))))**(k != 0)



"""Arc-Cosine."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Arccos(Dist):
    """
    Arc-Cosine.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Arccos(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Arccos(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.6435 0.9273 1.1593 1.3694]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6    0.8    0.9165 0.9798]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.2171 0.4843 1.5211 1.0265]
        >>> print(numpy.around(distribution.mom(1), 4))
        1.0
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[1.     0.8406 0.8083]
         [1.     0.1416 0.1492]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range() >= -1)
        assert numpy.all(dist.range() <= 1)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        return evaluation.evaluate_density(
            dist, numpy.cos(x), cache=cache)*numpy.sin(x)

    def _cdf(self, x, dist, cache):
        return 1-evaluation.evaluate_forward(dist, numpy.cos(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.arccos(evaluation.evaluate_inverse(dist, 1-q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.arccos(evaluation.evaluate_bound(
            dist, numpy.cos(x), cache=cache))[::-1]

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arccos(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.cos(dist)
        return self



"""Use of Bertran to calculate sparse segments."""
import numpy as np
import chaospy as cp


def sparse_segment(cords):
    r"""
    Create a segment of a sparse grid.

    Convert a ol-index to sparse grid coordinates on ``[0, 1]^N`` hyper-cube.
    A sparse grid of order ``D`` coencide with the set of sparse_segments where
    ``||cords||_1 <= D``.

    More specifically, a segment of:

    .. math::
        \cup_{cords \in C} sparse_segment(cords) == sparse_grid(M)

    where:

    .. math::
        C = {cords: M=sum(cords)}

    Args:
        cords (numpy.ndarray):
            The segment to extract. ``cord`` must consist of non-negative
            integers.

    Returns:
        Q (numpy.ndarray):
            Sparse segment where ``Q.shape==(K, sum(M))`` and ``K`` is segment
            specific.

    Examples:
        >>> print(cp.bertran.sparse_segment([0, 2]))
        [[0.5   0.125]
         [0.5   0.375]
         [0.5   0.625]
         [0.5   0.875]]
        >>> print(cp.bertran.sparse_segment([0, 1, 0, 0]))
        [[0.5  0.25 0.5  0.5 ]
         [0.5  0.75 0.5  0.5 ]]
    """
    cords = np.array(cords)+1
    slices = []
    for cord in cords:
        slices.append(slice(1, 2**cord+1, 2))

    grid = np.mgrid[slices]
    indices = grid.reshape(len(cords), np.prod(grid.shape[1:])).T
    sgrid = indices*2.**-cords
    return sgrid



"""Gumbel Distribution."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist


class gumbel(Archimedean):
    """Gumbel Distribution."""

    def __init__(self, length, theta=2., eps=1e-6):
        self.length = length
        Dist.__init__(self, th=float(theta), eps=eps)

    def __len__(self):
        return self.length

    def gen(self, x, th):
        return (-numpy.log(x))**th

    def igen(self, x, th):
        return numpy.e**(-x**th)


class Gumbel(Copula):
    r"""
    Gumbel Copula

    .. math::
        \phi(x;th) = \frac{x^{-th}-1}{th}
        \phi^{-1}(q;th) = (q*th + 1)^{-1/th}

    where `th` (or theta) is defined on the interval `[1,inf)`.

    Args:
        dist (Dist) : The Distribution to wrap
        theta (float) : Copula parameter

    Returns:
        (Dist) : The resulting copula distribution.

    Examples:
        >>> distribution = chaospy.Gumbel(
        ...     chaospy.Iid(chaospy.Uniform(), 2), theta=2)
        >>> print(distribution)
        Gumbel(Iid(Uniform(lower=0, upper=1), 2), theta=2)
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.5439 0.3319 0.2369]
          [0.6426 0.3802 0.2534]
          [0.7482 0.4292 0.2653]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[ 2.3224  4.6584 11.7092]
         [ 2.5896  5.4078 18.5493]
         [ 2.0433  4.601  23.5876]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.3151 0.665  0.1675 0.3745]]
    """

    def __init__(self, dist, theta, eps=1e-6):
        """
        Args:
            dist (Dist) : The Distribution to wrap
            theta (float) : Copula parameter
        """
        self._repr = {"theta": theta}
        Copula.__init__(self, dist=dist, trans=gumbel(len(dist), theta, eps))



"""
Perhaps the simplest method for creating orthogonal polynomials is
Gram-Schmidt's method. It consist of using projections to iteratively make new
terms in an expansion orthogonal to the previous ones.
"""
import logging

import numpy
import chaospy


def orth_gs(order, dist, normed=False, sort="G", cross_truncation=1., **kws):
    """
    Gram-Schmidt process for generating orthogonal polynomials.

    Args:
        order (int, Poly):
            The upper polynomial order. Alternative a custom polynomial basis
            can be used.
        dist (Dist):
            Weighting distribution(s) defining orthogonality.
        normed (bool):
            If True orthonormal polynomials will be used instead of monic.
        sort (str):
            Ordering argument passed to poly.basis. If custom basis is used,
            argument is ignored.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        (Poly):
            The orthogonal polynomial expansion.

    Examples:
        >>> Z = chaospy.J(chaospy.Normal(), chaospy.Normal())
        >>> print(chaospy.around(chaospy.orth_gs(2, Z), 4))
        [1.0, q1, q0, q1^2-1.0, q0q1, q0^2-1.0]
    """
    logger = logging.getLogger(__name__)
    dim = len(dist)

    if isinstance(order, int):
        if order == 0:
            return chaospy.poly.Poly(1, dim=dim)
        basis = chaospy.poly.basis(
            0, order, dim, sort, cross_truncation=cross_truncation)
    else:
        basis = order

    basis = list(basis)

    polynomials = [basis[0]]

    if normed:
        for idx in range(1, len(basis)):

            # orthogonalize polynomial:
            for idy in range(idx):
                orth = chaospy.descriptives.E(
                    basis[idx]*polynomials[idy], dist, **kws)
                basis[idx] = basis[idx] - polynomials[idy]*orth

            # normalize:
            norms = chaospy.descriptives.E(polynomials[-1]**2, dist, **kws)
            if norms <= 0:
                logger.warning("Warning: Polynomial cutoff at term %d", idx)
                break
            basis[idx] = basis[idx] / numpy.sqrt(norms)

            polynomials.append(basis[idx])

    else:

        norms = [1.]
        for idx in range(1, len(basis)):

            # orthogonalize polynomial:
            for idy in range(idx):
                orth = chaospy.descriptives.E(
                    basis[idx]*polynomials[idy], dist, **kws)
                basis[idx] = basis[idx] - polynomials[idy] * orth / norms[idy]

            norms.append(
                chaospy.descriptives.E(polynomials[-1]**2, dist, **kws))
            if norms[-1] <= 0:
                logger.warning("Warning: Polynomial cutoff at term %d", idx)
                break

            polynomials.append(basis[idx])

    return chaospy.poly.Poly(polynomials, dim=dim, shape=(len(polynomials),))



"""Sinus."""
import numpy
from ..baseclass import Dist
from .. import evaluation, approximation


class Sin(Dist):
    """
    Sinus.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Sin(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Sin(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.1987 0.3894 0.5646 0.7174]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.0203 1.0857 1.2116 1.4353]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.608  0.1148 0.8136 0.4637]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.4597
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.4597 0.4138 0.4192]
         [1.     0.0614 0.0466]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.arcsin(x), cache=cache)/numpy.sqrt(1-x*x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.arcsin(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.sin(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.sin(evaluation.evaluate_bound(
            dist, numpy.arcsin(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"




import numpy

from .. import distributions, poly as polynomials
from ..external import SampleDist


def QoI_Dist(poly, dist, sample=10000, **kws):
    """
    Constructs distributions for the quantity of interests.

    The function constructs a kernel density estimator (KDE) for each
    polynomial (poly) by sampling it.  With the KDEs, distributions (Dists) are
    constructed.  The Dists can be used for e.g. plotting probability density
    functions (PDF), or to make a second uncertainty quantification simulation
    with that newly generated Dists.

    Args:
        poly (Poly):
            Polynomial of interest.
        dist (Dist):
            Defines the space where the samples for the KDE is taken from the
            poly.
        sample (int):
            Number of samples used in estimation to construct the KDE.

    Returns:
        (numpy.ndarray):
            The constructed quantity of interest (QoI) distributions, where
            ``qoi_dists.shape==poly.shape``.

    Examples:
        >>> dist = chaospy.Normal(0, 1)
        >>> x = chaospy.variable(1)
        >>> poly = chaospy.Poly([x])
        >>> qoi_dist = chaospy.QoI_Dist(poly, dist)
        >>> values = qoi_dist[0].pdf([-0.75, 0., 0.75])
        >>> print(numpy.around(values, 8))
        [0.29143037 0.39931708 0.29536329]
    """
    shape = poly.shape
    poly = polynomials.flatten(poly)
    dim = len(dist)

    #sample from the inumpyut dist
    samples = dist.sample(sample, **kws)

    qoi_dists = []
    for i in range(0, len(poly)):
        #sample the polynomial solution
        if dim == 1:
            dataset = poly[i](samples)
        else:
            dataset = poly[i](*samples)

        lo = dataset.min()
        up = dataset.max()

        #creates qoi_dist
        qoi_dist = SampleDist(dataset, lo, up)
        qoi_dists.append(qoi_dist)

    #reshape the qoi_dists to match the shape of the inumpyut poly
    qoi_dists = numpy.array(qoi_dists, distributions.Dist)
    qoi_dists = qoi_dists.reshape(shape)

    if not shape:
        qoi_dists = qoi_dists.item()

    return qoi_dists



import logging
from functools import partial
import numpy

from . import evaluation
from .. import quadrature


def find_interior_point(
        distribution,
        parameters=None,
        cache=None,
        iterations=1000,
        retall=False,
        seed=None,
):
    """
    Find interior point of the distribution where forward evaluation is
    guarantied to be both ``distribution.fwd(xloc) > 0`` and
    ``distribution.fwd(xloc) < 1``.

    Args:
        distribution (Dist): Distribution to find interior on.
        parameters (Optional[Dict[Dist, numpy.ndarray]]): Parameters for the
            distribution.
        cache (Optional[Dict[Dist, numpy.ndarray]]): Memory cache for the
            location in the evaluation so far.
        iterations (int): The number of iterations allowed to be performed
        retall (bool): If provided, lower and upper bound which guaranties that
            ``distribution.fwd(lower) == 0`` and
            ``distribution.fwd(upper) == 1`` is returned as well.
        seed (Optional[int]): Fix random seed.

    Returns:
        numpy.ndarray: An input array with shape ``(len(distribution),)`` which
        is guarantied to be on the interior of the probability distribution.


    Example:
        >>> distribution = chaospy.MvNormal([1, 2, 3], numpy.eye(3)+.03)
        >>> midpoint, lower, upper = find_interior_point(
        ...     distribution, retall=True, seed=1234)
        >>> print(lower.T)
        [[-64. -64. -64.]]
        >>> print(numpy.around(midpoint, 4).T)
        [[  0.6784 -33.7687 -19.0182]]
        >>> print(upper.T)
        [[16. 16. 16.]]
        >>> distribution = chaospy.Uniform(1000, 1010)
        >>> midpoint, lower, upper = find_interior_point(
        ...     distribution, retall=True, seed=1234)
        >>> print(numpy.around(lower, 4))
        [[-1.]]
        >>> print(numpy.around(midpoint, 4))
        [[1009.8873]]
        >>> print(numpy.around(upper, 4))
        [[1024.]]
    """
    random_state = numpy.random.get_state()
    numpy.random.seed(seed)

    forward = partial(evaluation.evaluate_forward, cache=cache,
                      distribution=distribution, parameters=parameters)

    dim = len(distribution)
    upper = numpy.ones((dim, 1))
    for _ in range(100):
        indices = forward(x_data=upper) < 1
        if not numpy.any(indices):
            break
        upper[indices] *= 2

    lower = -numpy.ones((dim, 1))
    for _ in range(100):
        indices = forward(x_data=lower) > 0
        if not numpy.any(indices):
            break
        lower[indices] *= 2

    for _ in range(iterations):

        rand = numpy.random.random(dim)
        proposal = (rand*lower.T + (1-rand)*upper.T).T
        evals = forward(x_data=proposal)

        indices0 = evals > 0
        indices1 = evals < 1

        range_ = numpy.random.choice(dim, size=dim, replace=False)

        upper_ = numpy.where(indices1, upper, evals)
        for idx in range_:
            if upper.flatten()[idx] == upper_.flatten()[idx]:
                continue
            if numpy.all(forward(x_data=upper_) == 1):
                upper = upper_
                break
            upper_[idx] = upper[idx]

        lower_ = numpy.where(indices0, lower, evals)
        for idx in range_:
            if lower.flatten()[idx] == lower_.flatten()[idx]:
                continue
            if numpy.all(forward(x_data=lower_) == 0):
                lower = lower_
                break
            lower_[idx] = lower[idx]

        if numpy.all(indices0 & indices1):
            break

    else:
        if retall:
            return proposal, lower, upper
        return proposal
        raise evaluation.DependencyError(
            "Too many iterations required to find interior point.")

    numpy.random.set_state(random_state)
    if retall:
        return proposal, lower, upper
    return proposal


def approximate_inverse(
        distribution,
        qloc,
        parameters=None,
        cache=None,
        iterations=100,
        tol=1e-5,
        seed=None,
):
    """
    Calculate the approximation of the inverse Rosenblatt transformation.

    Uses forward Rosenblatt, probability density function (derivative) and
    boundary function to apply hybrid Newton-Raphson and binary search method.

    Args:
        distribution (Dist): Distribution to estimate inverse Rosenblatt.
        qloc (numpy.ndarray): Input values. All values must be on [0,1] and
            ``qloc.shape == (dim,size)`` where dim is the number of dimensions
            in distribution and size is the number of values to calculate
            simultaneously.
        parameters (Optional[Dict[Dist, numpy.ndarray]]): Parameters for the
            distribution.
        cache (Optional[Dict[Dist, numpy.ndarray]]): Memory cache for the
            location in the evaluation so far.
        iterations (int): The number of iterations allowed to be performed
        tol (float): Tolerance parameter determining convergence.
        seed (Optional[int]): Fix random seed.

    Returns:
        numpy.ndarray: Approximation of inverse Rosenblatt transformation.

    Example:
        >>> distribution = chaospy.Normal(1000, 10)
        >>> qloc = numpy.array([[0.1, 0.2, 0.9]])
        >>> print(numpy.around(approximate_inverse(
        ...     distribution, qloc, seed=1234), 4))
        [[ 987.1845  991.5838 1012.8155]]
        >>> print(numpy.around(distribution.inv(qloc), 4))
        [[ 987.1845  991.5838 1012.8155]]
    """
    logger = logging.getLogger(__name__)
    logger.debug("init approximate_inverse: %s", distribution)

    # lots of initial values:
    xloc, xlower, xupper = find_interior_point(
        distribution, cache=cache, parameters=parameters, retall=True, seed=seed)
    xloc = (xloc.T * numpy.zeros(qloc.shape).T).T
    xlower = (xlower.T + numpy.zeros(qloc.shape).T).T
    xupper = (xupper.T + numpy.zeros(qloc.shape).T).T
    uloc = numpy.zeros(qloc.shape)
    ulower = -qloc
    uupper = 1-qloc
    indices = numpy.ones(qloc.shape[-1], dtype=bool)

    for idx in range(2*iterations):

        # evaluate function:
        uloc[:, indices] = (evaluation.evaluate_forward(
            distribution, xloc, cache=cache, parameters=parameters)-qloc)[:, indices]

        # convergence criteria:
        indices[indices] = numpy.any(numpy.abs(xupper-xlower) > tol, 0)[indices]
        logger.debug(
            "iter: %s : %s : %s (%s)",
            numpy.mean(xlower, -1),
            numpy.mean(xloc, -1),
            numpy.mean(xupper, -1),
            numpy.mean(indices),
        )
        if not numpy.any(indices):
            break

        # narrow down lower boundary:
        ulower[:, indices] = numpy.where(uloc <= 0, uloc, ulower)[:, indices]
        xlower[:, indices] = numpy.where(uloc <= 0, xloc, xlower)[:, indices]

        # narrow down upper boundary:
        uupper[:, indices] = numpy.where(uloc >= 0, uloc, uupper)[:, indices]
        xupper[:, indices] = numpy.where(uloc >= 0, xloc, xupper)[:, indices]

        # Newton increment every second iteration:
        xloc_ = numpy.inf
        if idx % 2 == 0:
            derivative = evaluation.evaluate_density(
                distribution, xloc, cache=cache, parameters=parameters)[:, indices]
            derivative = numpy.where(derivative, derivative, numpy.inf)

            xloc_ = xloc[:, indices] - uloc[:, indices] / derivative

        # use binary search if Newton increment is outside bounds:
        xloc[:, indices] = numpy.where(
            (xloc_ < xupper[:, indices]) & (xloc_ > xlower[:, indices]),
            xloc_, 0.5*(xupper+xlower)[:, indices])

    else:
        logger.warning(
            "Too many iterations required to estimate inverse.")
        logger.info("{} out of {} did not converge.".format(
            numpy.sum(indices), len(indices)))
    logger.debug("end approximate_inverse: %s", distribution)
    return xloc


def approximate_moment(
        dist,
        K,
        retall=False,
        control_var=None,
        rule="F",
        order=1000,
        **kws
):
    """
    Approximation method for estimation of raw statistical moments.

    Args:
        dist : Dist
            Distribution domain with dim=len(dist)
        K : numpy.ndarray
            The exponents of the moments of interest with shape (dim,K).
        control_var : Dist
            If provided will be used as a control variable to try to reduce
            the error.
        acc (:py:data:typing.Optional[int]):
            The order of quadrature/MCI
        sparse : bool
            If True used Smolyak's sparse grid instead of normal tensor
            product grid in numerical integration.
        rule : str
            Quadrature rule
            Key     Description
            ----    -----------
            "G"     Optimal Gaussian quadrature from Golub-Welsch. Slow for
                    high order.
            "E"     Gauss-Legendre quadrature
            "C"     Clenshaw-Curtis quadrature. Exponential growth rule is
                    used when sparse is True to make the rule nested.

            Monte Carlo Integration
            Key     Description
            ----    -----------
            "H"     Halton sequence
            "K"     Korobov set
            "L"     Latin hypercube sampling
            "M"     Hammersley sequence
            "R"     (Pseudo-)Random sampling
            "S"     Sobol sequence

        antithetic (:py:data:typing.Optional[numpy.ndarray]):
            List of bool. Represents the axes to mirror using antithetic
            variable during MCI.
    """
    dim = len(dist)
    shape = K.shape
    size = int(K.size/dim)
    K = K.reshape(dim, size)

    if dim > 1:
        shape = shape[1:]

    X, W = quadrature.generate_quadrature(order, dist, rule=rule, **kws)

    grid = numpy.mgrid[:len(X[0]), :size]
    X = X.T[grid[0]].T
    K = K.T[grid[1]].T
    out = numpy.prod(X**K, 0)*W

    if control_var is not None:

        Y = control_var.ppf(dist.fwd(X))
        mu = control_var.mom(numpy.eye(len(control_var)))

        if (mu.size == 1) and (dim > 1):
            mu = mu.repeat(dim)

        for d in range(dim):
            alpha = numpy.cov(out, Y[d])[0, 1]/numpy.var(Y[d])
            out -= alpha*(Y[d]-mu)

    out = numpy.sum(out, -1)
    return out


def approximate_density(
        dist,
        xloc,
        parameters=None,
        cache=None,
        eps=1.e-7
):
    """
    Approximate the probability density function.

    Args:
        dist : Dist
            Distribution in question. May not be an advanced variable.
        xloc : numpy.ndarray
            Location coordinates. Requires that xloc.shape=(len(dist), K).
        eps : float
            Acceptable error level for the approximations
        retall : bool
            If True return Graph with the next calculation state with the
            approximation.

    Returns:
        numpy.ndarray: Local probability density function with
            ``out.shape == xloc.shape``. To calculate actual density function,
            evaluate ``numpy.prod(out, 0)``.

    Example:
        >>> distribution = chaospy.Normal(1000, 10)
        >>> xloc = numpy.array([[990, 1000, 1010]])
        >>> print(numpy.around(approximate_density(distribution, xloc), 4))
        [[0.0242 0.0399 0.0242]]
        >>> print(numpy.around(distribution.pdf(xloc), 4))
        [[0.0242 0.0399 0.0242]]
    """
    if parameters is None:
        parameters = dist.prm.copy()
    if cache is None:
        cache = {}

    xloc = numpy.asfarray(xloc)
    lo, up = numpy.min(xloc), numpy.max(xloc)
    mu = .5*(lo+up)
    eps = numpy.where(xloc < mu, eps, -eps)*xloc

    floc = evaluation.evaluate_forward(
        dist, xloc, parameters=parameters.copy(), cache=cache.copy())
    for d in range(len(dist)):
        xloc[d] += eps[d]
        tmp = evaluation.evaluate_forward(
            dist, xloc, parameters=parameters.copy(), cache=cache.copy())
        floc[d] -= tmp[d]
        xloc[d] -= eps[d]

    floc = numpy.abs(floc / eps)
    return floc



"""Student-T distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class student_t(Dist):
    """Student-T distribution."""

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        return special.gamma(.5*a+.5)*(1+x*x/a)**(-.5*a-.5) /\
                (numpy.sqrt(a*numpy.pi)*special.gamma(.5*a))

    def _cdf(self, x, a):
        return special.stdtr(a, x)

    def _ppf(self, q, a):
        return special.stdtrit(a, q)

    def _bnd(self, x, a):
        return self._ppf(1e-10, a), self._ppf(1-1e-10, a)

    def _mom(self, k, a):
        if numpy.any(a < k):
            raise ValueError("too high mom for student-t")
        out = special.gamma(.5*k+.5)* \
                special.gamma(.5*a-.5*k)*a**(.5*k)
        return numpy.where(k%2==0, out/(numpy.pi**.5*special.gamma(.5*a)), 0)

    def _ttr(self, k, a):
        return 0., k*a*(a-k+1.)/ ((a-2*k)*(a-2*k+2))


class StudentT(Add):
    """
    (Non-central) Student-t distribution.

    Args:
        df (float, Dist) : Degrees of freedom
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scale parameter

    Examples:
        >>> distribution = chaospy.StudentT(2, 2, 2)
        >>> print(distribution)
        StudentT(df=2, loc=2, scale=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [-0.1213  1.4226  2.5774  4.1213]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0905 0.1663 0.1663 0.0905]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 2.913  -1.4132  7.8594  1.8992]
        >>> print(distribution.mom(1))
        2.0
    """

    def __init__(self, df=1, loc=0, scale=1):
        self._repr = {"df": df, "loc": loc, "scale": scale}
        Add.__init__(self, left=student_t(df)*scale, right=loc)


Student_t = deprecation_warning(StudentT, "Student_t")



"""Bradford distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class bradford(Dist):
    """Standard Bradford distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return  c / (c*x + 1.0) / numpy.log(1.0+c)

    def _cdf(self, x, c):
        return numpy.log(1.0+c*x) / numpy.log(c+1.0)

    def _ppf(self, q, c):
        return ((1.0+c)**q-1)/c

    def _bnd(self, x, c):
        return 0, 1


class Bradford(Add):
    """
    Bradford distribution.

    Args:
        shape (float, Dist) : Shape parameter
        lower (float, Dist) : Location of lower threshold
        upper (float, Dist) : Location of upper threshold

    Examples:
        >>> distribution = chaospy.Bradford(0.8, 4, 6)
        >>> print(distribution)
        Bradford(lower=4, shape=0.8, upper=6)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [4.     4.3957 4.8541 5.385  6.    ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6805 0.5875 0.5072 0.4379 0.3781]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.171  4.1748 5.8704 4.8192]
        >>> print(numpy.around(distribution.mom(1), 4))
        4.9026
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[5.0195 5.0028 5.0009]
         [0.3314 0.2664 0.2571]]
    """
    def __init__(self, shape=1, lower=0, upper=1):
        self._repr = {"shape": shape, "lower": lower, "upper": upper}
        Add.__init__(self, left=bradford(shape)*(upper-lower), right=lower)



"""
Descriptives are a collection of statistical analysis tools that can be used to
analyse :ref:`distributions` and :ref:`polynomials`, both as an expansion (see
:ref:`orthogonality`) and as results (see :ref:`regression` and
:ref:`spectral`).  For example, the expected value operator
:func:`~chaospy.descriptives.expected.E` can be applied on distributions
directly as follows::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> expected = chaospy.E(distribution)
    >>> print(expected)
    0.5

For multivariate distributions::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 1),
    ...     chaospy.Normal(0, 1)
    ... )
    >>> expected = chaospy.E(distribution)
    >>> print(expected)
    [0.5 0. ]


For simple polynomials, distribution goes as the second argument. In other
words, it calculates the expected value of the unit variable with respect to
distribution. For example::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 1),
    ...     chaospy.Normal(0, 1)
    ... )
    >>> q0, q1 = chaospy.variable(2)
    >>> expected = chaospy.E(q1, distribution)
    >>> print(expected)
    0.0
"""
from .expected import E
from .conditional import E_cond
from .covariance import Cov
from .kurtosis import Kurt
from .percentile import Perc
from .quantity_of_interest import QoI_Dist
from .skewness import Skew
from .standard_deviation import Std
from .variance import Var

from .correlation import *
from .sensitivity import *



"""Log-laplace distribution."""
import numpy
from scipy import special, misc

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class log_laplace(Dist):
    """Log-laplace distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        cd2 = c/2.0
        c = numpy.where(x < 1, c, -c)
        return cd2*x**(c-1)

    def _cdf(self, x, c):
        return numpy.where(x < 1, 0.5*x**c, 1-0.5*x**(-c))

    def _ppf(self, q, c):
        return numpy.where(q < 0.5, (2.*q)**(1./c), (2*(1.-q))**(-1./c))

    def _bnd(self, x, c):
        return 0.0, self._ppf(1-1e-10, c)

class LogLaplace(Add):
    """
    Log-laplace distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.LogLaplace(2, 2, 2)
        >>> print(distribution)
        LogLaplace(scale=2, shape=2, shift=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [3.2649 3.7889 4.2361 5.1623]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.3162 0.4472 0.3578 0.1265]
        >>> print(numpy.around(distribution.sample(4), 4))
        [4.4028 2.9592 8.3425 3.9641]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=log_laplace(shape)*scale, right=shift)


Loglaplace = deprecation_warning(LogLaplace, "Loglaplace")



"""Generalized logistic type 1 distribution."""
import numpy
from scipy import special, misc

from ..baseclass import Dist
from ..operators.addition import Add

class logistic(Dist):
    """Generalized logistic type 1 distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return numpy.e**-x/(1+numpy.e**-x)**(c+1)

    def _cdf(self, x, c):
        return (1+numpy.e**-x)**-c

    def _ppf(self, q, c):
        return -numpy.log(q**(-1./c)-1)

    def _bnd(self, x, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class Logistic(Add):
    """
    Generalized logistic type 1 distribution
    Sech squared distribution

    Args:
        loc (float, Dist): Location parameter
        scale (float, Dist): Scale parameter
        skew (float, Dist): Shape parameter

    Examples:
        >>> distribution = chaospy.Logistic(2, 2, 2)
        >>> print(distribution)
        Logistic(loc=2, scale=2, skew=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.5761 3.0855 4.4689 6.2736]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0553 0.0735 0.0676 0.0422]
        >>> print(numpy.around(distribution.sample(4), 4))
        [4.8799 0.6656 9.3128 3.6415]
        >>> print(numpy.around(distribution.mom(1), 4))
        4.0
    """
    def __init__(self, loc=0, scale=1, skew=1):
        self._repr = {"loc": loc, "scale": scale, "skew": skew}
        Add.__init__(self, left=logistic(skew)*scale, right=loc)



"""Exponential Weibull distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class exponential_weibull(Dist):
    """Exponential Weibull distribution."""

    def __init__(self, a=1, c=1):
        Dist.__init__(self, a=a, c=c)

    def _pdf(self, x, a, c):
        exc = numpy.exp(-x**c)
        return a*c*(1-exc)**(a-1) * exc * x**(c-1)

    def _cdf(self, x, a, c):
        exm1c = -numpy.expm1(-x**c)
        return (exm1c)**a

    def _ppf(self, q, a, c):
        return (-numpy.log1p(-q**(1.0/a)))**(1.0/c)

    def _bnd(self, x, a, c):
        return 0, self._ppf(1-1e-10, a, c)


class ExponentialWeibull(Add):
    """
    Exponential Weibull distribution.

    Args:
        alpha (float, Dist): First shape parameter
        kappa (float, Dist): Second shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.ExponentialWeibull(2, 2, 2, 1)
        >>> print(distribution)
        ExponentialWeibull(alpha=2, kappa=2, scale=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.5398 3.0009 3.4412 3.9989]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.3807 0.4651 0.4262 0.2832]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.5711 2.2872 4.8376 3.1776]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.2916
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.731  4.1391 4.5115]
         [0.7486 1.4897 2.2135]]
    """
    def __init__(self, alpha=1, kappa=1, scale=1, shift=0):
        self._repr = {
            "alpha": alpha, "kappa": kappa, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=exponential_weibull(alpha, kappa)*scale, right=shift)


Exponweibull = deprecation_warning(ExponentialWeibull, "Exponweibull")



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['chaospy',
 'chaospy.bertran',
 'chaospy.chol',
 'chaospy.descriptives',
 'chaospy.descriptives.correlation',
 'chaospy.descriptives.sensitivity',
 'chaospy.distributions',
 'chaospy.distributions.collection',
 'chaospy.distributions.copulas',
 'chaospy.distributions.evaluation',
 'chaospy.distributions.operators',
 'chaospy.distributions.sampler',
 'chaospy.distributions.sampler.sequences',
 'chaospy.external',
 'chaospy.orthogonal',
 'chaospy.poly',
 'chaospy.poly.collection',
 'chaospy.poly.constructor',
 'chaospy.quadrature',
 'chaospy.quadrature.genz_keister',
 'chaospy.quadrature.recurrence']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'scipy']

setup_kwargs = {
    'name': 'chaospy',
    'version': '3.0.18',
    'description': 'Numerical tool for perfroming uncertainty quantification',
    'long_description': '.. image:: doc/.static/chaospy_logo.svg\n   :height: 300 px\n   :width: 300 px\n   :align: center\n\n|circleci| |codecov| |pypi| |readthedocs|\n\n.. |circleci| image:: https://circleci.com/gh/jonathf/chaospy/tree/master.svg?style=shield\n    :target: https://circleci.com/gh/jonathf/chaospy/tree/master\n.. |codecov| image:: https://codecov.io/gh/jonathf/chaospy/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jonathf/chaospy\n.. |pypi| image:: https://badge.fury.io/py/chaospy.svg\n    :target: https://badge.fury.io/py/chaospy\n.. |readthedocs| image:: https://readthedocs.org/projects/chaospy/badge/?version=master\n    :target: http://chaospy.readthedocs.io/en/master/?badge=master\n\nChaospy is a numerical tool for performing uncertainty quantification using\npolynomial chaos expansions and advanced Monte Carlo methods implemented in\nPython 2 and 3.\n\nIf you are using this software in work that will be published, please cite the\njournal article: `Chaospy: An open source tool for designing methods of\nuncertainty quantification <http://dx.doi.org/10.1016/j.jocs.2015.08.008>`_\n\n.. contents:: Table of Contents:\n\nInstallation\n------------\n\nInstallation should be straight forward::\n\n    pip install chaospy\n\nAnd you should be ready to go.\n\nAlternatively, to get the most current experimental version, the code can be\ninstalled from Github as follows::\n\n    git clone git@github.com:jonathf/chaospy.git\n    cd chaospy\n    git checkout <tag or branch of interest>\n    pip install .\n\nExample Usage\n-------------\n\n``chaospy`` is created to be simple and modular. A simple script to implement\npoint collocation method will look as follows:\n\n.. code-block:: python\n\n    import chaospy\n    import numpy\n\n    # your code wrapper goes here\n    def foo(coord, prm):\n        """Function to do uncertainty quantification on."""\n        return prm[0] * numpy.e ** (-prm[1] * numpy.linspace(0, 10, 100))\n\n    # bi-variate probability distribution\n    distribution = chaospy.J(chaospy.Uniform(1, 2), chaospy.Uniform(0.1, 0.2))\n\n    # polynomial chaos expansion\n    polynomial_expansion = chaospy.orth_ttr(8, distribution)\n\n    # samples:\n    samples = distribution.sample(1000)\n\n    # evaluations:\n    evals = [foo(sample) for sample in samples.T]\n\n    # polynomial approximation\n    foo_approx = chaospy.fit_regression(\n        polynomial_expansion, samples, evals)\n\n    # statistical metrics\n    expected = chaospy.E(foo_approx, distribution)\n    deviation = chaospy.Std(foo_approx, distribution)\n\nFor a more extensive description of what going on, see the `tutorial\n<https://chaospy.readthedocs.io/en/master/tutorial.html>`_.\n\nFor a collection of recipes, see the `cookbook\n<https://chaospy.readthedocs.io/en/master/cookbook.html>`_.\n\nRelated Projects\n----------------\n\nChaospy is being used in other related projects that requires uncertainty\nquantification components ``chaospy`` provides.\n\n+-----------------+-----------------------------------------------------------+\n| `easyVVUQ`_     | Library designed to facilitate verification, validation   |\n|                 | and uncertainty quantification.                           |\n+-----------------+-----------------------------------------------------------+\n| `STARFiSh`_     | Shell-based, scientific simulation program                |\n|                 | for blood flow in mammals.                                |\n+-----------------+-----------------------------------------------------------+\n| `SurUQ`_        | Surrogate modeling with nested uncertainty quantification.|\n+-----------------+-----------------------------------------------------------+\n| `UncertainPy`_  | Uncertainty quantification and sensitivity analysis,      |\n|                 | tailored towards computational neuroscience.              |\n+-----------------+-----------------------------------------------------------+\n\n.. _easyVVUQ: https://github.com/UCL-CCS/EasyVVUQ\n.. _STARFiSh: https://www.ntnu.no/starfish\n.. _SurUQ: https://github.com/redmod-team/suruq\n.. _UncertainPy: https://github.com/simetenn/uncertainpy\n\nAlso a few shout-outs:\n\n+--------------+--------------------------------------------------------------+\n| `OpenTURNS`_ | Thanks to `R\xc3\xa9gis Lebrun`_ for both proposing a collaboration |\n|              | and creating an initial implementation of both               |\n|              | `Chaospy Compatibility`_ in `OpenTURNS`_ and                 |\n|              | `OpenTURNS Compatibility`_ in ``chaospy``.                   |\n+--------------+--------------------------------------------------------------+\n| `orthopy`_   | Thanks to `Nico Schl\xc3\xb6mer`_ for providing the implementation  |\n| `quadpy`_    | for several of the quadrature integration methods.           |\n+--------------+--------------------------------------------------------------+\n| ``UQRF``     | Thanks to `Florian K\xc3\xbcnzner`_ for providing the               |\n|              | implementation for `sample distribution`_.                   |\n+--------------+--------------------------------------------------------------+\n\n.. _OpenTURNS: http://openturns.github.io/openturns/latest\n.. _R\xc3\xa9gis Lebrun: https://github.com/regislebrun\n.. _Chaospy Compatibility: http://openturns.github.io/openturns/latest/user_manual/_generated/openturns.ChaospyDistribution.html\n.. _OpenTURNS Compatibility: https://chaospy.readthedocs.io/en/master/recipes/external.html#module-chaospy.external.openturns_\n.. _orthopy: https://github.com/nschloe/orthopy\n.. _quadpy: https://github.com/nschloe/quadpy\n.. _Nico Schl\xc3\xb6mer: https://github.com/nschloe\n.. _Florian K\xc3\xbcnzner: https://github.com/flo2k\n.. _sample distribution: https://chaospy.readthedocs.io/en/master/recipes/external.html#module-chaospy.external.samples\n\nQuestions & Troubleshooting\n---------------------------\n\nFor any problems and questions you might have related to ``chaospy``, please\nfeel free to file an `issue <https://github.com/jonathf/chaospy/issues>`_.\n',
    'author': 'Jonathan Feinberg',
    'author_email': None,
    'url': 'https://github.com/jonathf/chaospy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



"""Pearson's correlation matrix."""
import numpy
from scipy.stats import spearmanr

from ... import distributions, poly as polynomials
from ..covariance import Cov


def Corr(poly, dist=None, **kws):
    """
    Correlation matrix of a distribution or polynomial.

    Args:
        poly (Poly, Dist):
            Input to take correlation on. Must have ``len(poly)>=2``.
        dist (Dist):
            Defines the space the correlation is taken on.  It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Correlation matrix with
            ``correlation.shape == poly.shape+poly.shape``.

    Examples:
        >>> Z = chaospy.MvNormal([3, 4], [[2, .5], [.5, 1]])
        >>> print(numpy.around(chaospy.Corr(Z), 4))
        [[1.     0.3536]
         [0.3536 1.    ]]

        >>> x = chaospy.variable()
        >>> Z = chaospy.Normal()
        >>> print(numpy.around(chaospy.Corr([x, x**2], Z), 4))
        [[1. 0.]
         [0. 1.]]
    """
    if isinstance(poly, distributions.Dist):
        poly, dist = polynomials.variable(len(poly)), poly
    else:
        poly = polynomials.Poly(poly)

    if not poly.shape:
        return numpy.ones((1, 1))

    cov = Cov(poly, dist, **kws)
    var = numpy.diag(cov)
    vvar = numpy.sqrt(numpy.outer(var, var))
    return numpy.where(vvar > 0, cov/vvar, 0)



"""Tools for creating multi-indices"""
from __future__ import division

import itertools
import numpy


def bindex(start, stop=None, dim=1, sort="G", cross_truncation=1.):
    """
    Generator for creating multi-indices.

    Args:
        start (Union[int, numpy.ndarray]):
            The lower order of the indices. If array of int, counts as lower
            bound for each axis.
        stop (Union[int, numpy.ndarray, None]):
            the maximum shape included. If omitted: stop <- start; start <- 0
            If int is provided, set as largest total order. If array of int,
            set as upper bound for each axis.
        dim (int):
            The number of dimensions in the expansion
        sort (str):
            Criteria to sort the indices by.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. Ignored if ``stop`` is a array.

    Returns:
        list:
            Order list of indices.

    Examples:
        >>> print(bindex(2, 3, 2).tolist())
        [[0, 2], [1, 1], [2, 0], [0, 3], [1, 2], [2, 1], [3, 0]]
        >>> print(bindex(2, [1, 3], 2, cross_truncation=0).tolist())
        [[0, 2], [1, 1], [0, 3], [1, 2], [1, 3]]
        >>> print(bindex([1, 2], [2, 3], 2, cross_truncation=0).tolist())
        [[1, 2], [1, 3], [2, 2], [2, 3]]
        >>> print(bindex(1, 3, 2, cross_truncation=0).tolist())  # doctest: +NORMALIZE_WHITESPACE
        [[0, 1], [1, 0], [0, 2], [1, 1], [2, 0], [0, 3], [1, 2], [2, 1], [3, 0],
            [1, 3], [2, 2], [3, 1], [2, 3], [3, 2], [3, 3]]
        >>> print(bindex(1, 3, 2, cross_truncation=1).tolist())
        [[0, 1], [1, 0], [0, 2], [1, 1], [2, 0], [0, 3], [1, 2], [2, 1], [3, 0]]
        >>> print(bindex(1, 3, 2, cross_truncation=1.5).tolist())
        [[0, 1], [1, 0], [0, 2], [1, 1], [2, 0], [0, 3], [3, 0]]
        >>> print(bindex(1, 3, 2, cross_truncation=2).tolist())
        [[0, 1], [1, 0], [0, 2], [2, 0], [0, 3], [3, 0]]
        >>> print(bindex(0, 1, 3).tolist())
        [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]]
        >>> print(bindex([1, 1], 3, 2, cross_truncation=0).tolist())
        [[1, 1], [1, 2], [2, 1], [1, 3], [2, 2], [3, 1], [2, 3], [3, 2], [3, 3]]
    """
    if stop is None:
        start, stop = 0, start
    start = numpy.array(start, dtype=int).flatten()
    stop = numpy.array(stop, dtype=int).flatten()
    sort = sort.upper()
    start[start < 0] = 0

    indices = _bindex(start, stop, dim, cross_truncation)
    if "G" in sort:
        indices = indices[numpy.lexsort([numpy.sum(indices, -1)])]

    if "I" in sort:
        indices = indices[::-1]

    if "R" in sort:
        indices = indices[:, ::-1]

    return indices


def _bindex(start, stop, dim=1, cross_truncation=1.):
    # At the beginning the current list of indices just ranges over the
    # last dimension.
    bound = stop.max()+1
    range_ = numpy.arange(bound, dtype=int)
    indices = range_[:, numpy.newaxis]

    for idx in range(dim-1):

        # Repeats the current set of indices.
        # e.g. [0,1,2] -> [0,1,2,0,1,2,...,0,1,2]
        indices = numpy.tile(indices, (bound, 1))

        # Stretches ranges over the new dimension.
        # e.g. [0,1,2] -> [0,0,...,0,1,1,...,1,2,2,...,2]
        front = range_.repeat(len(indices)//bound)[:, numpy.newaxis]

        # Puts them two together.
        indices = numpy.column_stack((front, indices))

        # Truncate at each iteration to ensure memory usage is low enough
        if stop.size == 1 and cross_truncation > 0:
            lhs = numpy.sum(indices**(1/cross_truncation), -1)
            rhs = numpy.max(stop, -1)**(1/cross_truncation)
            indices = indices[lhs <= rhs]
        else:
            indices = indices[numpy.all(indices <= stop, -1)]

    if start.size == 1:
        indices = indices[numpy.sum(indices, -1) >= start.item()]
    else:
        indices = indices[numpy.all(indices >= start, -1)]
    return indices.reshape(-1, dim)
